#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
from typing import Dict, Any


from ..pdi.constants import PdiCommand, IrdaAction
from ..pdi.irda_req import IrdaSequence, IrdaReq
from ..protocol.constants import Direction, CommandScope
from .component_state import LcsState, P, log, SCOPE_TO_STATE_MAP


class IrdaState(LcsState):
    """
    Maintain the state of a Sensor Track (Irda)
    """

    def __init__(self, scope: CommandScope = CommandScope.IRDA) -> None:
        if scope != CommandScope.IRDA:
            raise ValueError(f"Invalid scope: {scope}, expected {CommandScope.IRDA.name}")
        super().__init__(scope)
        self._sequence: IrdaSequence | None = None
        self._loco_rl: int | None = 255
        self._loco_lr: int | None = 255
        self._last_train_id = self._last_engine_id = self._last_dir = self._product_id = None

    def __repr__(self) -> str:
        if self.sequence and self.sequence != IrdaSequence.NONE:
            rle = f"{self._loco_rl}" if self._loco_rl and self._loco_rl != 255 else "Any"
            lre = f"{self._loco_lr}" if self._loco_lr and self._loco_lr != 255 else "Any"
            rl = f" When Engine ID (R -> L): {rle}"
            lr = f" When Engine ID (L -> R): {lre}"
        else:
            rl = lr = ""
        le = f" Last Engine ID: {self._last_engine_id}" if self._last_engine_id else ""
        lt = f" Last Train ID: {self._last_train_id}" if self._last_train_id else ""
        if self._last_dir is not None:
            ld = " L --> R" if self._last_dir == 1 else " R --> L"
        else:
            ld = ""
        return f"Sensor Track {self.address}: Sequence: {self.sequence_str}{rl}{lr}{le}{lt}{ld}"

    def update(self, command: P) -> None:
        from ..comm.comm_buffer import CommBuffer
        from .component_state_store import ComponentStateStore

        if command:
            with self._cv:
                self._is_known = True
                super().update(command)
                if isinstance(command, IrdaReq) and command.pdi_command == PdiCommand.IRDA_RX:
                    if command.action == IrdaAction.CONFIG:
                        self._sequence = command.sequence
                        self._loco_rl = command.loco_rl
                        self._loco_lr = command.loco_lr
                    elif command.action == IrdaAction.SEQUENCE:
                        self._sequence = command.sequence
                    elif command.action == IrdaAction.DATA:
                        # change engine/train speed, based on the direction of travel
                        self._last_engine_id = command.engine_id
                        self._last_train_id = command.train_id
                        self._last_dir = command.direction
                        self._product_id = command.product_id
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"IRDA {self.address} Sequence: {self.sequence} Command: {command}")
                        if (
                            self.sequence
                            in {
                                IrdaSequence.SLOW_SPEED_NORMAL_SPEED,
                                IrdaSequence.NORMAL_SPEED_SLOW_SPEED,
                            }
                            and CommBuffer.is_server()
                        ):
                            rr_speed = None
                            if command.is_right_to_left:
                                rr_speed = "slow" if self.sequence == IrdaSequence.SLOW_SPEED_NORMAL_SPEED else "normal"
                            elif command.is_left_to_right:
                                rr_speed = "normal" if self.sequence == IrdaSequence.SLOW_SPEED_NORMAL_SPEED else "slow"
                            if rr_speed:
                                address = None
                                scope = CommandScope.ENGINE
                                if command.train_id:
                                    address = command.train_id
                                    scope = CommandScope.TRAIN
                                elif command.engine_id:
                                    address = command.engine_id
                                state = ComponentStateStore.get_state(scope, address)
                                if state is not None:
                                    from ..protocol.sequence.ramped_speed_req import RampedSpeedReq

                                    # noinspection PyTypeChecker
                                    RampedSpeedReq(address, rr_speed, scope=scope, is_tmcc=state.is_tmcc).send()
                            # send update to Train and component engines as well
                            orig_scope = command.scope
                            orig_tmcc_id = command.tmcc_id
                            try:
                                if command.engine_id:
                                    engine_state = ComponentStateStore.get_state(CommandScope.ENGINE, command.engine_id)
                                    command.scope = CommandScope.ENGINE
                                    command.tmcc_id = command.engine_id
                                    engine_state.update(command)
                            finally:
                                command.scope = orig_scope
                                command.tmcc_id = orig_tmcc_id
                    self.changed.set()
                    self._cv.notify_all()

    @property
    def sequence(self) -> IrdaSequence:
        return self._sequence

    @property
    def sequence_str(self) -> str | None:
        return self.sequence.name.title() if self.sequence else "NA"

    @property
    def last_direction(self) -> Direction:
        if self._last_dir == 1:
            return Direction.L2R
        elif self._last_dir == 0:
            return Direction.R2L
        else:
            return Direction.UNKNOWN

    @property
    def is_left_to_right(self) -> bool:
        return self.last_direction == 1

    @property
    def is_right_to_left(self) -> bool:
        return self.last_direction == 0

    @property
    def last_engine_id(self) -> int:
        return self._last_engine_id

    @property
    def last_train_id(self) -> int:
        return self._last_train_id

    @property
    def is_engine(self) -> bool:
        return (self.is_train is False) and (self._last_engine_id is not None) and (self._last_engine_id > 0)

    @property
    def product_type(self) -> str:
        return self._product_id

    @property
    def is_train(self) -> bool:
        return (self._last_train_id is not None) and (self._last_train_id > 0)

    def as_bytes(self) -> bytes:
        # TODO: return IrdaAction.DATA
        if self.is_known:
            return IrdaReq(
                self.address,
                PdiCommand.IRDA_RX,
                IrdaAction.CONFIG,
                sequence=self._sequence,
                loco_rl=self._loco_rl,
                loco_lr=self._loco_lr,
            ).as_bytes
        else:
            return bytes()

    def as_dict(self) -> Dict[str, Any]:
        d = super()._as_dict()
        d["last_direction"] = self.last_direction.name.lower()
        d["last_engine_id"] = self.last_engine_id
        d["last_train_id"] = self.last_train_id
        d["sequence"] = self.sequence.name.lower() if self.sequence else "none"
        return d


SCOPE_TO_STATE_MAP.update({CommandScope.IRDA: IrdaState})



import re
from unittest import mock

# noinspection PyPackageRequirements
import pytest

from src.pytrain.comm.comm_buffer import CommBufferSingleton, CommBuffer
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import DEFAULT_BAUDRATE, DEFAULT_PORT
from src.pytrain.protocol.multibyte.multibyte_constants import *
from src.pytrain.protocol.tmcc1.tmcc1_constants import *
from src.pytrain.protocol.tmcc2.tmcc2_constants import *
from ..test_base import TestBase


# noinspection PyMethodMayBeStatic
class TestCommandReq(TestBase):
    def teardown_method(self, test_method):
        super().teardown_method(test_method)
        CommBuffer.build().shutdown()

    def test_send_command(self):
        with mock.patch.object(CommandReq, "_enqueue_command") as mk_enqueue_command:
            # test TMCC1 commands, beginning with HALT
            CommandReq.send_request(TMCC1HaltCommandEnum.HALT)
            mk_enqueue_command.assert_called_once_with(
                0xFEFFFF.to_bytes(3, byteorder="big"),
                1,
                0,
                0,
                DEFAULT_BAUDRATE,
                DEFAULT_PORT,
                None,
            )
            mk_enqueue_command.reset_mock()

            # Route command
            CommandReq.send_request(TMCC1RouteCommandEnum.FIRE, 10)
            mk_enqueue_command.assert_called_once_with(
                0xFED51F.to_bytes(3, byteorder="big"),
                1,
                0,
                0,
                DEFAULT_BAUDRATE,
                DEFAULT_PORT,
                None,
            )
            mk_enqueue_command.reset_mock()

            # test all TMCC1 command defs
            address = 23
            for cdef in [TMCC1AuxCommandEnum, TMCC1SwitchCommandEnum, TMCC1EngineCommandEnum]:
                for cmd in cdef:
                    if cmd in [TMCC1EngineCommandEnum.RELATIVE_SPEED, TMCC1AuxCommandEnum.RELATIVE_SPEED]:
                        continue  # can't test defs that map data, yet
                    data = self.generate_random_data(cmd)
                    CommandReq.send_request(cmd, address, data)
                    bits = cmd.command_def.bits | (int.from_bytes(cmd.command_def.first_byte, byteorder="big") << 16)
                    bits |= address << 7
                    if data != 0:
                        bits |= data
                    mk_enqueue_command.assert_called_once_with(
                        bits.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
                    )
                    mk_enqueue_command.reset_mock()

            # test engine defs again with TRAIN scope
            address = 1
            for cmd in TMCC1EngineCommandEnum:
                if cmd == TMCC1EngineCommandEnum.RELATIVE_SPEED:
                    continue  # can't test defs that map data, yet
                data = self.generate_random_data(cmd)
                CommandReq.send_request(cmd, address, data, CommandScope.TRAIN)
                bits = cmd.command_def.bits
                bits |= address << 7
                bits &= TMCC1_TRAIN_COMMAND_PURIFIER
                bits |= TMCC1_TRAIN_COMMAND_MODIFIER
                bits |= int.from_bytes(cmd.command_def.first_byte, byteorder="big") << 16
                if cmd.command_def.num_data_bits > 0:
                    bits |= data
                mk_enqueue_command.assert_called_once_with(
                    bits.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
                )
                mk_enqueue_command.reset_mock()

            # random switch command
            CommandReq.send_request(TMCC1SwitchCommandEnum.THRU, 15)
            mk_enqueue_command.assert_called_once_with(
                0xFE4780.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
            )
            mk_enqueue_command.reset_mock()

            # random acc command
            CommandReq.send_request(TMCC1AuxCommandEnum.AUX2_OPT_ONE, 15)
            mk_enqueue_command.assert_called_once_with(
                0xFE878D.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
            )
            mk_enqueue_command.reset_mock()

            # random engine command
            CommandReq.send_request(TMCC1EngineCommandEnum.RELATIVE_SPEED, 28, -5)
            mk_enqueue_command.assert_called_once_with(
                0xFE0E40.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
            )
            mk_enqueue_command.reset_mock()

            # test TMCC2 commands
            for cdef in [TMCC2RouteCommandEnum, TMCC2EngineCommandEnum]:
                for cmd in cdef:
                    if cmd == TMCC2EngineCommandEnum.RELATIVE_SPEED:
                        continue  # can't test defs that map data, yet
                    data = self.generate_random_data(cmd)
                    bits = cmd.command_def.bits
                    bits |= address << 9
                    bits |= int.from_bytes(cmd.command_def.first_byte, byteorder="big") << 16
                    CommandReq.send_request(cmd, address, data)
                    if cmd.command_def.num_data_bits > 0:
                        bits |= data
                    mk_enqueue_command.assert_called_once_with(
                        bits.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
                    )
                    mk_enqueue_command.reset_mock()

            CommandReq.send_request(TMCC2RouteCommandEnum.FIRE, 10)
            mk_enqueue_command.assert_called_once_with(
                0xFA14FD.to_bytes(3, byteorder="big"), 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None
            )
            mk_enqueue_command.reset_mock()

    def test__determine_first_byte(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                assert isinstance(cmd.value, CommandDef)
                assert cmd.value.first_byte == CommandReq._determine_first_byte(cmd.value, cmd.value.scope)

        # retest for Trains
        for cdef in self.all_command_enums:
            for cmd in cdef:
                req = self.build_request(cmd, 1, 2, CommandScope.TRAIN)
                cmd_bytes = req.as_bytes
                fb = cmd_bytes[0].to_bytes(1, byteorder="big")
                assert fb == CommandReq._determine_first_byte(cmd.value, CommandScope.TRAIN)

    def test__vet_request(self):
        # all command defs should pass
        for cdef in self.all_command_enums:
            for cmd in cdef:
                CommandReq._vet_request(cmd, 1, 0, CommandScope.ENGINE)

        # test that non-CommandDefEnums fail
        with pytest.raises(TypeError, match="Command def not recognized: 'invalid_command'"):
            # noinspection PyTypeChecker
            CommandReq._vet_request("invalid_command", 1, 0, CommandScope.ENGINE)

    def test__enqueue_command(self):
        with mock.patch.object(CommBufferSingleton, "enqueue_command") as mk_enqueue_command:
            # test _enqueue_command with byte string
            CommandReq._enqueue_command(b"\x01\x02\x03", 1, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None)
            mk_enqueue_command.assert_called_once_with(b"\x01\x02\x03", 0)
            mk_enqueue_command.reset_mock()

            # test repeat argument
            CommandReq._enqueue_command(b"\x01\x02\x03", 5, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None)
            assert mk_enqueue_command.call_count == 5
            mk_enqueue_command.reset_mock()

        # test for invalid arguments
        with pytest.raises(ValueError, match=re.escape("repeat must be equal to or greater than 1 (-5)")):
            CommandReq._enqueue_command(b"\x01\x02\x03", -5, 0, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None)
        with pytest.raises(ValueError, match=re.escape("delay must be equal to or greater than 0 (-6)")):
            CommandReq._enqueue_command(b"\x01\x02\x03", 1, -6, 0, DEFAULT_BAUDRATE, DEFAULT_PORT, None)

    def test_address(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                address = self.generate_random_address(cmd)
                req = self.build_request(cmd, address, 2)
                if req.command_def.is_addressable:
                    assert req.address == address
                else:
                    assert req.address == DEFAULT_ADDRESS

    def test_data(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                data = self.generate_random_data(cmd)
                req = self.build_request(cmd, 1, data)
                assert req.data == data

    def test_scope(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                address = self.generate_random_address(cmd)
                data = self.generate_random_data(cmd)
                req = self.build_request(cmd, address, data)
                assert req.scope == cmd.scope

    def test_command_def(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                req = self.build_request(cmd)
                assert req.command_def == cmd.command_def

    def test_bits(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                address = self.generate_random_address(cmd)
                data = self.generate_random_data(cmd)
                req = self.build_request(cmd, address, data)
                if isinstance(cmd, TMCC2MultiByteEnum):
                    pass
                else:
                    # make sure all bits in the definition are also in the request
                    # the other bits in the request are the address and data values
                    assert req.bits & cmd.value.bits == cmd.value.bits

    def test_num_data_bits(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                req = self.build_request(cmd, 1, 2, CommandScope.TRAIN)
                assert req.num_data_bits == cmd.command_def.num_data_bits

    def test_syntax(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                req = self.build_request(cmd)
                assert req.syntax == cmd.syntax

    def test_identifier(self):
        for cdef in self.all_command_enums:
            for cmd in cdef:
                req = self.build_request(cmd)
                assert req.identifier == cmd.value.identifier

    # noinspection DuplicatedCode
    def test_build_tmcc1_command_req(self):
        """
        Build all the TMCC1 CommandReqs and verify that their command bytes
        map back to the sane request
        """
        for tmcc_enums in [
            TMCC1HaltCommandEnum,
            TMCC1SwitchCommandEnum,
            TMCC1AuxCommandEnum,
            TMCC1RouteCommandEnum,
            TMCC1EngineCommandEnum,
        ]:
            for tmcc_enum in tmcc_enums:
                if tmcc_enum.command_def.is_data:
                    n_times = 10
                else:
                    n_times = 1
                if tmcc_enums == TMCC1EngineCommandEnum:
                    scopes = [None, CommandScope.TRAIN]
                else:
                    scopes = [None]
                for scope in scopes:
                    for _ in range(n_times):
                        req = self.build_request(tmcc_enum, scope=scope)
                        # do reverse lookup
                        req_from_bytes = CommandReq.from_bytes(req.as_bytes)
                        if tmcc_enum.command_def.is_alias:
                            # if the enum is an alias for another command,
                            # check results against that command_def
                            alias_enum = tmcc_enum.command_def.alias
                            if isinstance(alias_enum, tuple):
                                alias_enum = alias_enum[0]
                                alias_data = tmcc_enum.command_def.alias[1]
                            else:
                                alias_data = None
                            assert req_from_bytes.command == alias_enum
                            assert req_from_bytes.command_def == alias_enum.command_def
                            assert req_from_bytes.num_data_bits == alias_enum.command_def.num_data_bits
                            assert alias_enum.command_def.is_valid_data(req_from_bytes.data)
                            assert alias_enum.command_def.is_valid_data(alias_data)
                        else:
                            assert req_from_bytes.command == req.command
                            assert req_from_bytes.command_def == req.command_def
                            assert req_from_bytes.num_data_bits == req.num_data_bits
                            assert req_from_bytes.data == req.data
                        assert req_from_bytes.address == req.address
                        assert req_from_bytes.syntax == req.syntax
                        assert req_from_bytes.identifier == req.identifier
                        assert req_from_bytes.scope == req.scope
                        assert req_from_bytes.is_tmcc1 == req.is_tmcc1
                        assert req_from_bytes.is_tmcc2 == req.is_tmcc2
                        assert req_from_bytes.as_bytes == req.as_bytes

    # noinspection DuplicatedCode
    def test_build_tmcc2_command_req(self):
        """
        Build all the TMCC2 CommandReqs and verify that their command bytes
        map back to the sane request
        """
        for tmcc_enums in [TMCC2HaltCommandEnum, TMCC2EngineCommandEnum, TMCC2RouteCommandEnum]:
            for tmcc_enum in tmcc_enums:
                if tmcc_enum.command_def.is_data:
                    n_times = 10
                else:
                    n_times = 1
                if tmcc_enums == TMCC2EngineCommandEnum:
                    scopes = [None, CommandScope.TRAIN]
                else:
                    scopes = [None]
                for scope in scopes:
                    for _ in range(n_times):
                        req = self.build_request(tmcc_enum, scope=scope)
                        # do reverse lookup
                        req_from_bytes = CommandReq.from_bytes(req.as_bytes)
                        if tmcc_enum.command_def.is_alias:
                            # if the enum is an alias for another command,
                            # check results against that command_def
                            alias_enum = tmcc_enum.command_def.alias
                            if isinstance(alias_enum, tuple):
                                alias_data = alias_enum[1]
                                alias_enum = alias_enum[0]
                            else:
                                alias_data = None
                            assert req_from_bytes.command == alias_enum
                            assert req_from_bytes.command_def == alias_enum.command_def
                            assert req_from_bytes.num_data_bits == alias_enum.command_def.num_data_bits
                            assert alias_enum.command_def.is_data is False or alias_enum.command_def.is_valid_data(
                                alias_data
                            )
                        else:
                            assert req_from_bytes.command == req.command
                            assert req_from_bytes.command_def == req.command_def
                            assert req_from_bytes.num_data_bits == req.num_data_bits
                            assert req_from_bytes.data == req.data
                        assert req_from_bytes.address == req.address
                        assert req_from_bytes.syntax == req.syntax
                        assert req_from_bytes.identifier == req.identifier
                        assert req_from_bytes.scope == req.scope
                        assert req_from_bytes.is_tmcc1 == req.is_tmcc1
                        assert req_from_bytes.is_tmcc2 == req.is_tmcc2
                        assert req_from_bytes.as_bytes == req.as_bytes

    # noinspection DuplicatedCode
    def test_build_parameter_command_req(self):
        """
        Build all the Parameter CommandReqs and verify that their command bytes
        map back to the sane request
        """
        scopes = [None, CommandScope.TRAIN]
        for tmcc_enums in [
            TMCC2RailSoundsDialogControl,
            TMCC2RailSoundsEffectsControl,
            TMCC2EffectsControl,
            TMCC2LightingControl,
            TMCC2MultiByteEnum,
        ]:
            for tmcc_enum in tmcc_enums:
                for scope in scopes:
                    req = self.build_request(tmcc_enum, scope=scope)
                    # do reverse lookup
                    req_from_bytes = CommandReq.from_bytes(req.as_bytes)
                    assert req_from_bytes.command == req.command
                    assert req_from_bytes.command_def == req.command_def
                    assert req_from_bytes.num_data_bits == req.num_data_bits
                    assert req_from_bytes.data == req.data
                    assert req_from_bytes.address == req.address
                    assert req_from_bytes.syntax == req.syntax
                    assert req_from_bytes.identifier == req.identifier
                    assert req_from_bytes.scope == req.scope
                    assert req_from_bytes.is_tmcc1 == req.is_tmcc1
                    assert req_from_bytes.is_tmcc2 == req.is_tmcc2
                    assert req_from_bytes.as_bytes == req.as_bytes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

# -------------------------------------------------------------------------------
#                                                                               -
#  Python dual-logging setup (console and log file),                            -
#  supporting different log levels and colorized output                         -
#                                                                               -
#  Created by Fonic <https://github.com/fonic>                                  -
#  Date: 04/05/20 - 02/07/23                                                    -
#                                                                               -
#  Based on:                                                                    -
#  https://stackoverflow.com/a/13733863/1976617                                 -
#  https://uran198.github.io/en/python/2016/07/12/colorful-python-logging.html  -
#  https://en.wikipedia.org/wiki/ANSI_escape_code#Colors                        -
#                                                                               -
# -------------------------------------------------------------------------------

import logging
import logging.handlers
import os

# Imports
import sys


# Logging formatter supporting colorized output
class LogFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.CRITICAL: "\033[1;35m",  # bright/bold magenta
        logging.ERROR: "\033[1;31m",  # bright/bold red
        logging.WARNING: "\033[1;33m",  # bright/bold yellow
        logging.INFO: "\033[40;37m",  # white / light gray
        logging.DEBUG: "\033[1;30m",  # bright/bold dark gray
    }

    RESET_CODE = "\033[0m"

    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if self.color is True and record.levelno in self.COLOR_CODES:
            record.color = self.COLOR_CODES[record.levelno]
            record.no_color = self.RESET_CODE
        else:
            record.color = ""
            record.no_color = ""
        return super(LogFormatter, self).format(record)


class ConsoleFormatter(LogFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def formatException(self, exc_info):
        return ""

    def formatStack(self, stack_trace):
        return ""


# Set up logging
def set_up_logging(
    console_log_output: str = "stdout",
    console_log_level: str = "INFO",
    console_log_color: bool = True,
    console_template: str = "%(color)s%(message)s%(no_color)s",
    logfile_file: str = "pytrain.log",
    logfile_log_level: str = "INFO",
    logfile_log_color: bool = False,
    logfile_template: str = "%(color)s[%(asctime)s] [%(name)s] [%(levelname)-8s] %(message)s%(no_color)s",
) -> bool:
    # Create logger
    # For simplicity, we use the root logger, i.e. call 'logging.getLogger()'
    # without name argument. This way we can simply use module methods for
    # logging throughout the script. An alternative would be exporting the
    # logger, i.e. 'global logger; logger = logging.getLogger("<name>")'
    logger = logging.getLogger()

    # Set global log level to 'debug' (required for handler levels to work)
    logger.setLevel(logging.DEBUG)

    # Create console handler
    console_log_output = console_log_output.lower()
    if console_log_output == "stdout":
        console_log_output = sys.stdout
    elif console_log_output == "stderr":
        console_log_output = sys.stderr
    else:
        print("Failed to set console output: invalid output: '%s'" % console_log_output)
        return False
    console_handler = logging.StreamHandler(console_log_output)

    # Set console log level
    try:
        console_handler.setLevel(console_log_level.upper())  # only accepts uppercase level names
    except Exception as e:
        print(f"Failed to set console log level: invalid level: '{console_log_level}': {e}")
        return False

    # Create and set formatter, add console handler to logger
    console_formatter = ConsoleFormatter(fmt=console_template, color=console_log_color)
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    # Create log file handler
    try:
        should_roll_over = os.path.isfile(logfile_file)
        logfile_handler = logging.handlers.RotatingFileHandler(logfile_file, backupCount=4)
        if should_roll_over:  # log already exists, roll over!
            logfile_handler.doRollover()
    except Exception as exception:
        print(f"Failed to set up log file: {str(exception)}")
        return False

    # Set log file log level
    try:
        logfile_handler.setLevel(logfile_log_level.upper())  # only accepts uppercase level names
    except Exception as e:
        print(f"Failed to set log file log level: invalid level: '{logfile_log_level}': {e}")
        return False

    # Create and set formatter, add log file handler to logger
    logfile_formatter = LogFormatter(fmt=logfile_template, color=logfile_log_color)
    logfile_handler.setFormatter(logfile_formatter)
    logger.addHandler(logfile_handler)

    # Success
    return True


# Main function
def main():
    # Set up logging
    if not set_up_logging():
        print("Failed to set up logging, aborting.")
        return 1

    # Log some messages
    logging.debug("Debug message")
    logging.info("Info message")
    logging.warning("Warning message")
    logging.error("Error message")
    logging.critical("Critical message")


# Call main function
if __name__ == "__main__":
    sys.exit(main())



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from ..utils.validations import Validations
from .gpio_device import GpioDevice, P


class SmokeFluidLoader(GpioDevice):
    def __init__(
        self,
        address: int,
        boom_left_pin: P,
        boom_right_pin: P,
        dispense_pin: P,
        lights_on_pin: P = None,
        lights_off_pin: P = None,
        command_control: bool = True,
        boom_rotary_encoder: bool = False,
        repeat_every: float = 0.02,
    ) -> None:
        Validations.validate_float(repeat_every, 0.005, 1, label="repeat_every")
        if command_control is True:
            if boom_rotary_encoder is True:
                from .py_rotary_encoder import PyRotaryEncoder

                self.boom_left_btn = self.boom_right_btn = None
                cmd = CommandReq.build(TMCC1AuxCommandEnum.RELATIVE_SPEED, address, data=0, scope=CommandScope.ACC)
                self.boom_re = PyRotaryEncoder(
                    boom_left_pin,
                    boom_right_pin,
                    cmd,
                    wrap=False,
                    initial_step=0,
                    max_steps=180,
                    steps_to_data=self.fast_step_to_data,
                    pause_for=repeat_every,
                    reset_after_motion=True,
                )
            else:
                # use momentary contact switch to rotate cab
                self.cab_re = None
                left_cmd, self.boom_left_btn, _ = self.make_button(
                    boom_left_pin,
                    command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                    address=address,
                    data=-1,
                    scope=CommandScope.ACC,
                    hold_repeat=True,
                    hold_time=repeat_every,
                )
                self.boom_left_btn.when_pressed = left_cmd.as_action()
                self.boom_left_btn.when_held = left_cmd.as_action()

                right_cmd, self.boom_right_btn, _ = self.make_button(
                    boom_right_pin,
                    command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                    address=address,
                    data=1,
                    scope=CommandScope.ACC,
                    hold_repeat=True,
                    hold_time=repeat_every,
                )
                self.boom_right_btn.when_pressed = right_cmd.as_action()
                self.boom_right_btn.when_held = right_cmd.as_action()

            dispense_req, self.dispense_btn, _ = self.make_button(
                dispense_pin,
                TMCC1AuxCommandEnum.BRAKE,
                address,
            )
            self.dispense_btn.when_pressed = dispense_req.as_action(repeat=2)

            if lights_on_pin:
                lights_on_req, self.lights_on_btn, _ = self.make_button(
                    lights_on_pin,
                    TMCC1AuxCommandEnum.NUMERIC,
                    address,
                    data=9,
                )
                self.lights_on_btn.when_pressed = lights_on_req.as_action(repeat=2)
            else:
                self.lights_on_btn = None

            if lights_off_pin:
                lights_off_req, self.lights_off_btn, _ = self.make_button(
                    lights_off_pin,
                    TMCC1AuxCommandEnum.NUMERIC,
                    address,
                    data=8,
                )
                self.lights_off_btn.when_pressed = lights_off_req.as_action(repeat=2)
            else:
                self.lights_off_btn = None
        else:
            raise NotImplementedError



from .tmcc1_command import TMCC1Command
from .tmcc1_constants import TMCC1SwitchCommandEnum

from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT


class SwitchCmd(TMCC1Command):
    def __init__(
        self,
        switch: int,
        state: TMCC1SwitchCommandEnum = TMCC1SwitchCommandEnum.THRU,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        if switch < 1 or switch > 99:
            raise ValueError("Switch must be between 1 and 99")
        req = CommandReq(state, switch)
        super().__init__(state, req, switch, 0, None, baudrate, port, server)
        self._command = self._build_command()



"""
Simple examples of how to associate Lionel commands to Raspberry Pi buttons
"""

from src.pytrain import Controller

"""
Build a train controller component that uses an I2C-connected Keypad and LCD.
"""
Controller.build(
    keypad_address=0x20,
    speed_pins=[17, 27],
    halt_pin=10,
    reset_pin=11,
    horn_pin=5,
    bell_pin=9,
    start_up_pin=21,
    shutdown_pin=20,
    boost_pin=15,
    brake_pin=14,
    fwd_pin=16,
    rev_pin=12,
    rpm_up_pin=8,
    rpm_down_pin=7,
    front_coupler_pin=24,
    rear_coupler_pin=25,
    vol_up_pin=26,
    vol_down_pin=19,
    tower_dialog_pin=13,
    engr_dialog_pin=6,
    quilling_horn_chn=0,
    base_online_pin=23,
    base_offline_pin=4,
    lcd_address=0x27,
    lcd_rows=4,
    lcd_cols=20,
)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from .gpio_device import GpioDevice, P


class PowerWatcher(GpioDevice):
    """
    Controls an LED that indicates power status of the attached Raspberry Pi.

    The PowerWatcher class is designed to manage a single LED that remains
    illuminated as long as the connected Raspberry Pi has power. It creates
    an LED instance configured with the specified GPIO pin and optionally
    sets its cathode configuration. The LED is initialized to indicate an
    "always-on" state and is cached for later control.

    Attributes
    ----------
    power_led : LED
        The initialized LED object indicating power status.

    Parameters
    ----------
    power_on_pin : P
        The GPIO pin used to control the power LED.
    cathode : bool, optional
        Specifies if the LED is cathode-based (default is True).
    """

    def __init__(
        self,
        power_on_pin: P,
        cathode: bool = True,
    ) -> None:
        """
        Illuminates a LED as long as the attached Pi has power
        """
        self.power_led = self.make_led(power_on_pin, cathode=cathode)
        self.power_led.value = 1
        self.cache_device(self.power_led)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import collections
import logging
import threading
from collections import defaultdict
from typing import List, Set, Tuple, TypeVar, Generic

from ..comm.comm_buffer import CommBuffer
from ..comm.command_listener import CommandListener, Message, Subscriber, Topic
from ..db.client_state_listener import ClientStateListener
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import BROADCAST_ADDRESS, CommandScope
from ..protocol.multibyte.multibyte_constants import TMCC2EffectsControl as Effects
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum as Aux
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum as Engine1
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum as Switch
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum as Engine2
from .component_state import SCOPE_TO_STATE_MAP, ComponentState, ComponentStateDict, SystemStateDict

log = logging.getLogger(__name__)

T = TypeVar("T", bound=ComponentState)
C = TypeVar("C", bound=CommandDefEnum)
E = TypeVar("E", bound=CommandDefEnum)


class ComponentStateStore:
    _instance: ComponentStateStore = None
    _lock = threading.RLock()

    @classmethod
    def build(
        cls,
        topics: List[str | CommandScope] = None,
        listeners: Tuple = None,
        is_base: bool = False,
        is_ser2: bool = False,
    ) -> ComponentStateStore:
        """
        Factory method to create a ComponentStateStore instance
        """
        return ComponentStateStore(topics, listeners, is_base=is_base, is_ser2=is_ser2)

    @classmethod
    def get(cls) -> ComponentStateStore:
        if cls._instance is None:
            raise AttributeError("ComponentStateStore not built")
        return cls._instance

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def get_state(cls, scope: CommandScope, address: int, create: bool = True) -> T | None:
        if cls._instance is None:
            raise AttributeError("ComponentStateStore not built")
        if create:
            return cls._instance._state[scope][address]
        else:
            return cls._instance.query(scope, address)

    @classmethod
    def set_state(cls, scope: CommandScope, address: int, state: T) -> None:
        if cls._instance is None:
            raise AttributeError("ComponentStateStore not built")
        cls._instance._state[scope][address] = state

    @classmethod
    def reset(cls) -> None:
        with cls._lock:
            if cls._instance:
                cls._instance._state.clear()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if ComponentStateStore._instance is None:
                ComponentStateStore._instance = super(ComponentStateStore, cls).__new__(cls)
                ComponentStateStore._instance._initialized = False
            return ComponentStateStore._instance

    def __init__(
        self,
        topics: List[str | CommandScope] = None,
        listeners: Tuple = None,
        is_base: bool = False,
        is_ser2: bool = False,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        self._dependencies = DependencyCache.build()
        self._listeners = listeners
        self._state: dict[CommandScope, ComponentStateDict] = SystemStateDict()
        self._is_base = is_base
        self._is_ser2 = is_ser2
        self._filter_updates = is_base is True and is_ser2 is True
        if topics:
            for topic in topics:
                if self.is_valid_topic(topic):
                    for listener in self._listeners:
                        listener.listen_for(self, topic)
                else:
                    raise TypeError(f"Topic {topic} is not valid")
        log.debug(self)

    def __repr__(self) -> str:
        return f"<ComponentStateStore Base: {self._is_base} Ser2: {self._is_ser2} Filtering: {self._filter_updates}>"

    def __contains__(self, scope: CommandScope) -> bool:
        return scope in self._state

    def __call__(self, command: Message) -> None:
        """
        Callback, per the Subscriber protocol in CommandListener
        """
        if command:
            if isinstance(command, CommandReq):
                if command.is_halt:  # send to all known devices
                    if self._filter_updates and command.is_filtered:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"Command {command} is suppressed")
                    else:
                        for scope in self._state:
                            for address in self._state[scope]:
                                self._state[scope][address].update(command)
                        return
                elif command.is_system_halt:  # send to all known engines and trains
                    for scope in self._state:
                        if scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                            for address in self._state[scope]:
                                self._state[scope][address].update(command)
                    return
            if command.scope in SCOPE_TO_STATE_MAP:
                if self._filter_updates and command.is_filtered:
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug(f"Command {command} is suppressed")
                else:
                    if command.address == BROADCAST_ADDRESS and command.scope != CommandScope.SYNC:  # broadcast address
                        for address in self._state[command.scope]:
                            self._state[command.scope][address].update(command)
                    else:  # update the device state (identified by scope/address)
                        # make sure we haven't filled a slot based on a previous engine's road number
                        if command.address > 99:
                            state = self.query(command.scope, command.address)
                            if state and state.address != command.address:
                                del self._state[command.scope][command.address]
                        self._state[command.scope][command.address].update(command)
            else:
                log.warning(f"Received Unknown State Update: {command.scope} {command}")

    @property
    def is_empty(self) -> bool:
        return len(self._state) == 0

    @property
    def is_base(self) -> bool:
        return self._is_base

    @property
    def is_ser2(self) -> bool:
        return self._is_ser2

    @property
    def is_filter_updates(self) -> bool:
        return self._filter_updates

    @staticmethod
    def is_valid_topic(topic: Topic) -> bool:
        return isinstance(topic, CommandScope) or (
            isinstance(topic, tuple) and len(topic) > 1 and isinstance(topic[0], CommandScope)
        )

    def listen_for(self, topics: Topic | List[Topic]) -> None:
        if isinstance(topics, list):
            for topic in topics:
                if self.is_valid_topic(topic):
                    for listener in self._listeners:
                        listener.listen_for(self, topic)
        else:
            if self.is_valid_topic(topics):
                for listener in self._listeners:
                    listener.listen_for(self, topics)

    def query(self, scope: CommandScope, address: int = None) -> T | List[T] | None:
        if scope in self._state:
            if address is None:
                return self.get_all(scope)
            elif address in self._state[scope]:
                return self._state[scope][address]
        return None

    def get_all(self, scope: CommandScope) -> List[T]:
        if scope in self._state:
            # ignore dups where we store an entry for the item's road number
            valids = {v.address for k, v in self._state[scope].items()}
            states = [v for k, v in self._state[scope].items() if k in valids]
            states.sort(key=lambda x: x.address)
            return states
        else:
            return []

    def keys(self, scope: CommandScope = None) -> List[CommandScope] | List[int]:
        if scope is None:
            li = list(self._state.keys())
            li.sort(key=lambda x: x.label)
        elif scope in self._state:
            states = self.get_all(scope)
            li = list(s.address for s in states)
            li.sort()
        else:
            li = []
        return li

    def component(self, scope: CommandScope, address: int) -> T:
        if (
            (
                scope
                in {CommandScope.ACC, CommandScope.TRAIN, CommandScope.SWITCH, CommandScope.IRDA, CommandScope.ROUTE}
                and 1 <= address <= 99
            )
            or (scope == CommandScope.ENGINE and 1 <= address <= 9999)
            or (scope == CommandScope.BASE and address == 0)
        ):
            return self._state[scope][address]
        raise ValueError(f"Invalid scope/address: {scope} {address}")

    def scopes(self) -> Set[CommandScope]:
        return set(self._state.keys())

    def addresses(self, scope: CommandScope) -> collections.Iterable[int]:
        return set(self._state[scope].keys())


# noinspection DuplicatedCode
class DependencyCache(Generic[T, C, E]):
    """
    Manages relationships between TMCC Commands. For example, sending the Reset command
    (number 0) to an engine causes the following results:
        - speed is set to zero
        - direction is set to Fwd
        - bell is disabled
        - engine rpm set to 0
        - engine labor set to 12
        - engine starts up immediately
    The reverse mapping, results to causes, is used to maintain a consistent state on a
    control panel. For example, if an indicator light specifies an engine is set to Rev,
    sending the reset command would turn that light off.
    """

    _instance: DependencyCache = None
    _lock = threading.RLock()

    @classmethod
    def build(cls) -> DependencyCache:
        """
        Factory method to create a ComponentStateStore instance
        """
        return DependencyCache()

    @classmethod
    def listen_for_enablers(cls, request: CommandReq, callback: Subscriber) -> List[E | Tuple[E, int]] | None:
        enablers = None
        if cls._instance is not None:
            if CommBuffer.is_server():
                listener = CommandListener.build()
            elif CommBuffer.is_client():
                listener = ClientStateListener.build()
            else:
                raise AttributeError("CommBuffer must be server or client")
            enablers = cls._instance.enabled_by(request.command, dereference_aliases=True, include_aliases=False)
            for enabler in enablers:
                if isinstance(enabler, tuple):
                    listener.listen_for(callback, request.scope, request.address, enabler[0], enabler[1])
                    listener.listen_for(callback, request.scope, BROADCAST_ADDRESS, enabler[0], enabler[1])
                else:
                    listener.listen_for(callback, request.scope, request.address, enabler)
                    listener.listen_for(callback, request.scope, BROADCAST_ADDRESS, enabler)
        return enablers

    @classmethod
    def listen_for_disablers(cls, request: CommandReq, callback: Subscriber) -> List[E | Tuple[E, int]] | None:
        disablers = None
        if cls._instance is not None:
            if CommBuffer.is_server():
                listener = CommandListener.build()
            elif CommBuffer.is_client():
                listener = ClientStateListener.build()
            else:
                raise AttributeError("CommBuffer must be server or client")
            disablers = cls._instance.disabled_by(request.command, dereference_aliases=True, include_aliases=False)
            for disabler in disablers:
                if isinstance(disabler, tuple):
                    listener.listen_for(callback, request.scope, request.address, disabler[0], disabler[1])
                    listener.listen_for(callback, request.scope, BROADCAST_ADDRESS, disabler[0], disabler[1])
                else:
                    listener.listen_for(callback, request.scope, request.address, disabler)
                    listener.listen_for(callback, request.scope, BROADCAST_ADDRESS, disabler)
        return disablers

    def __init__(self) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        self._causes: dict[C, set[E]] = defaultdict(set)
        self._caused_bys: dict[E, Set[C]] = defaultdict(set)
        self._toggles: dict[C, set[E]] = defaultdict(set)
        self._toggled_by: dict[E, set[C]] = defaultdict(set)
        self.initialize()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if DependencyCache._instance is None:
                DependencyCache._instance = super(DependencyCache, cls).__new__(cls)
                DependencyCache._instance._initialized = False
            return DependencyCache._instance

    @staticmethod
    def _harvest_commands(commands: Set[E], dereference_aliases: bool, include_aliases: bool) -> Set[E]:
        cmd_set = set()
        for cmd in commands:
            if isinstance(cmd, CommandDefEnum) and cmd.is_alias:
                if include_aliases is True:
                    cmd_set.add(cmd)
                if dereference_aliases is True:
                    cmd_set.add(cmd.alias)
            else:
                cmd_set.add(cmd)
        return cmd_set

    def causes(self, cause: C, *results: E) -> None:
        """
        Define the results that are triggered when a cause command occurs
        """
        for result in results:
            self._causes[cause].add(result)
            self._caused_bys[result].add(cause)
        # if the cause is an aliased command, register the base command
        if isinstance(cause, CommandDefEnum) and cause.is_alias and hasattr(cause, "alias") and cause.alias is not None:
            self.causes(cause.alias, *results)

    def caused_by(self, command: E, dereference_aliases: bool = False, include_aliases: bool = True) -> Set[C]:
        """
        Returns a list of CommandDefEnums that can cause the given command to be issued.
        These commands should be "listened for" to as indicators for this state change
        """
        if command in self._caused_bys:
            causes = self._harvest_commands(self._caused_bys[command], dereference_aliases, include_aliases)
            if command not in causes:
                # noinspection PyTypeChecker
                causes.update(self._harvest_commands({command}, dereference_aliases, include_aliases))
            return causes
        else:
            return {command}

    def results_in(self, command: C, dereference_aliases: bool = False, include_aliases: bool = True) -> Set[E]:
        """
        Returns a list of the CommandDefEnums that result from issuing the given command.
        """
        if command in self._causes:
            results = self._harvest_commands(self._causes[command], dereference_aliases, include_aliases)
            if command not in results:
                # noinspection PyTypeChecker
                results.update(self._harvest_commands({command}, dereference_aliases, include_aliases))
            return results
        elif isinstance(command, CommandDefEnum) and command.is_alias and command.alias in self._causes:
            results = self._harvest_commands(self._causes[command.alias], dereference_aliases, include_aliases)
            if command.is_alias not in results:
                results.update(self._harvest_commands({command.alias}, dereference_aliases, include_aliases))
            return results
        else:
            return {command}

    def enabled_by(
        self, command: C, dereference_aliases: bool = False, include_aliases: bool = True
    ) -> List[E | Tuple[E, int]]:
        return list(self._harvest_commands(self.caused_by(command), dereference_aliases, include_aliases))

    def disabled_by(
        self, command: C, dereference_aliases: bool = False, include_aliases: bool = True
    ) -> List[E | Tuple[E, int]]:
        disabled = set()
        if command in self._toggles:
            disabled.update(self._harvest_commands(self._toggles[command], dereference_aliases, include_aliases))
            for state in list(disabled):  # as we may add items, we need to loop over as copy
                disabled.update(self._harvest_commands(self.caused_by(state), dereference_aliases, include_aliases))
        return list(disabled)

    def toggles(self, actor: C, *toggles: E) -> None:
        for toggled in toggles:
            self._toggles[actor].add(toggled)
            self._toggled_by[toggled].add(actor)

    def initialize(self) -> None:
        self._causes.clear()
        self._caused_bys.clear()

        # define command relationships
        # self.causes(Halt1.HALT, Engine1.SPEED_STOP_HOLD, Engine2.SPEED_STOP_HOLD, Aux.AUX2_OFF, Aux.AUX1_OFF)
        # self.causes(Halt2.HALT, Engine2.SPEED_STOP_HOLD)
        self.causes(Engine2.SYSTEM_HALT, Engine2.SPEED_STOP_HOLD)

        # Engine commands, starting with Reset (Number 0)
        self.causes(
            Engine2.RESET,
            Engine2.STOP_IMMEDIATE,
            Engine2.FORWARD_DIRECTION,
            Engine2.START_UP_IMMEDIATE,
            Engine2.BELL_OFF,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
            Engine2.NUMERIC,
        )
        self.causes(
            Engine2.STOP_IMMEDIATE,
            Engine2.SPEED_STOP_HOLD,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
        )
        self.causes(
            Engine2.FORWARD_DIRECTION,
            Engine2.STOP_IMMEDIATE,
            Engine2.SPEED_STOP_HOLD,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
        )
        self.causes(
            Engine2.REVERSE_DIRECTION,
            Engine2.STOP_IMMEDIATE,
            Engine2.SPEED_STOP_HOLD,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
        )
        self.causes(
            Engine2.TOGGLE_DIRECTION,
            Engine2.STOP_IMMEDIATE,
            Engine2.SPEED_STOP_HOLD,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
        )

        self.causes(
            Engine2.SHUTDOWN_IMMEDIATE,
            Engine2.SPEED_STOP_HOLD,
            Engine2.STOP_IMMEDIATE,
            Engine2.DIESEL_RPM,
            Engine2.ENGINE_LABOR_DEFAULT,
        )

        # TMCC1 Engine Reset
        self.causes(
            Engine1.RESET,
            Engine1.SPEED_STOP_HOLD,
            Engine1.FORWARD_DIRECTION,
            Engine1.RPM_DOWN,
            Engine1.NUMERIC,
        )
        self.causes(Engine1.FORWARD_DIRECTION, Engine1.SPEED_STOP_HOLD)
        self.causes(Engine1.REVERSE_DIRECTION, Engine1.SPEED_STOP_HOLD)
        self.causes(Engine1.TOGGLE_DIRECTION, Engine1.SPEED_STOP_HOLD)

        # define command toggles; commands that are essentially mutually exclusive
        self.toggles(Switch.OUT, Switch.THRU)
        self.toggles(Switch.THRU, Switch.OUT)

        self.toggles(Aux.AUX1_OPT_ONE, Aux.AUX2_OPT_ONE)
        self.toggles(Aux.AUX2_OPT_ONE, Aux.AUX1_OPT_ONE)

        self.toggles(Engine2.FORWARD_DIRECTION, Engine2.REVERSE_DIRECTION)
        self.toggles(Engine2.REVERSE_DIRECTION, Engine2.FORWARD_DIRECTION)

        self.toggles(Effects.SMOKE_OFF, Effects.SMOKE_LOW, Effects.SMOKE_MEDIUM, Effects.SMOKE_HIGH)
        self.toggles(Effects.SMOKE_LOW, Effects.SMOKE_OFF, Effects.SMOKE_MEDIUM, Effects.SMOKE_HIGH)
        self.toggles(Effects.SMOKE_MEDIUM, Effects.SMOKE_LOW, Effects.SMOKE_OFF, Effects.SMOKE_HIGH)
        self.toggles(Effects.SMOKE_HIGH, Effects.SMOKE_LOW, Effects.SMOKE_MEDIUM, Effects.SMOKE_OFF)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from __future__ import annotations

import atexit
from threading import Thread, Event, RLock
from time import time

from .gpio_device import GpioDevice
from .i2c.oled import OledDevice, Oled
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum, TMCC1HaltCommandEnum
from ..comm.command_listener import CommandDispatcher
from ..protocol.command_req import CommandReq
from ..db.engine_state import EngineState
from ..db.component_state_store import ComponentStateStore
from ..db.state_watcher import StateWatcher
from ..protocol.constants import PROGRAM_NAME, CommandScope


class LaunchStatus(Thread, GpioDevice):
    def __init__(
        self,
        tmcc_id: int | EngineState = 39,
        title: str | None = None,
        address: int = 0x3C,
        device: OledDevice | str = OledDevice.ssd1309,
    ) -> None:
        self._lock = RLock()
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Launch Pad Status")

        if isinstance(tmcc_id, EngineState):
            self._monitored_state = tmcc_id
            self._tmcc_id = tmcc_id.address
            self._scope = tmcc_id.scope
        elif isinstance(tmcc_id, int) and 1 <= tmcc_id <= 99:
            self._tmcc_id = tmcc_id
            self._scope = CommandScope.ENGINE
            self._monitored_state = None
        else:
            raise ValueError(f"Invalid tmcc_id: {tmcc_id}")

        self._state_store = ComponentStateStore.get()
        self._dispatcher = CommandDispatcher.get()
        self._is_running = True
        self._ev = Event()
        self._state_watcher = None
        self._countdown: int | None = None
        self._holding = False
        self._aborted = False
        self._countdown_thread = None
        self._last_cmd = None
        self._last_cmd_time = time()
        self._hidden = False

        self._title = title if title else f"Pad {tmcc_id}"
        self._oled = Oled(address, device, auto_update=False)
        self.display.write(self.title, 0, center=True)
        self.display.write("T Minus  --:--", 1, center=True)

        # check for state synchronization
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        if self._sync_state and self._sync_state.is_synchronized is True:
            self._sync_watcher = None
            self.on_sync()
        else:
            self.update_display()
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)
        atexit.register(self.close)

    def __call__(self, cmd: CommandReq) -> None:
        try:
            if cmd == self._last_cmd and time() - self._last_cmd_time < 0.5:
                # ignore dupe commands, if received within 0.5 seconds of one another
                return
            last_cmd = self._last_cmd.command if self._last_cmd else None
            if cmd.command == TMCC1EngineCommandEnum.REAR_COUPLER:
                # launch in 15 seconds
                self.launch(15)
            elif cmd.command == TMCC1EngineCommandEnum.FRONT_COUPLER:
                # launch now
                self.launch(0)
            elif cmd.command == TMCC1EngineCommandEnum.NUMERIC:
                if cmd.data == 0:
                    # Num 0: Abort
                    self.abort()
                elif cmd.data == 5:
                    if (
                        last_cmd == TMCC1EngineCommandEnum.AUX1_OPTION_ONE
                        or self.aborted is True
                        or self.countdown is None
                    ):
                        # Aux 1/Num 5: Shutdown
                        self.recycle(False)
                        self.countdown = None
                        self._hide()
                    else:
                        # Num 5: Abort
                        self.abort()
                elif cmd.data == 3 and last_cmd == TMCC1EngineCommandEnum.AUX1_OPTION_ONE:
                    # Aux 1/Num 3: Startup
                    self.countdown = None
                    self._show()
            elif cmd.command == TMCC1HaltCommandEnum.HALT:
                self.abort()
        finally:
            self._last_cmd_time = time()
            self._last_cmd = cmd

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, value: str) -> None:
        self._title = value
        self.display.write(value, 0, center=True)
        self.update_display(clear=True)

    @property
    def countdown(self) -> int | None:
        return self._countdown

    @countdown.setter
    def countdown(self, value: int) -> None:
        with self._lock:
            self._countdown = value
            if value is None:
                r1 = "T Minus  --:--"
                self.recycle(False)
            else:
                if value < 0:
                    value = abs(value)
                    prefix = "T Minus -"
                else:
                    prefix = "Launch  +"
                minute = value // 60
                second = value % 60
                r1 = f"{prefix}{minute:02d}:{second:02d}"
            self.display.write(r1, 1, center=True)
            self.update_display()

    @property
    def display(self) -> Oled:
        return self._oled

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @property
    def is_synchronized(self) -> bool:
        return self._synchronized

    @property
    def state(self) -> EngineState:
        return self._monitored_state

    @property
    def holding(self) -> bool:
        return self._holding

    @property
    def aborted(self) -> bool:
        return self._aborted

    def launch(self, countdown: int = -30) -> None:
        with self._lock:
            self.recycle(False)
            if self._countdown_thread:
                self._countdown_thread.reset()
                self._countdown_thread = None
            if countdown is None:
                self.countdown = None
            else:
                countdown = countdown if countdown < 0 else -countdown
                self.countdown = countdown
                self._countdown_thread = CountdownThread(self, countdown)

    def abort(self) -> None:
        with self._lock:
            if self.countdown is None:
                return
            self._holding = False
            self._aborted = True
            if self._countdown_thread:
                self._countdown_thread.reset()
                self._countdown_thread = None
            self.display.write("** Abort **", 3, center=True, blink=True)
        self.update_display()

    def hold(self) -> None:
        with self._lock:
            if self.countdown is None or self.countdown > 0:
                return
            if self._countdown_thread:
                self._countdown_thread.hold()
            self.display.write("** Hold **", 3, center=True, blink=True)
            self._holding = True
        self.update_display()

    def resume(self) -> None:
        with self._lock:
            if self._holding is False:
                return
            self.display[3] = ""
            if self._countdown_thread:
                self._countdown_thread.resume()
            self._holding = False
        self.update_display()

    def update_display(self, clear: bool = False) -> None:
        with self._lock:
            self._show()
            if clear is True:
                self.display.clear()
                self.display.write(self.title, 0, center=True)
                self.display.write("T Minus  --:--", 1, center=True)
            # self.display.refresh_display()
            self.display.force_display()

    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True
            if self._monitored_state is None and self.tmcc_id and self.tmcc_id != 99:
                self._monitored_state = self._state_store.get_state(CommandScope.ENGINE, self.tmcc_id)
            self._dispatcher.subscribe(self, CommandScope.ENGINE, self.tmcc_id)
            self.update_display(clear=True)
            self._hide()

    def reset(self) -> None:
        with self._lock:
            self._dispatcher.unsubscribe(self, CommandScope.ENGINE, self.tmcc_id)
            self._holding = self._aborted = False
            if self._countdown_thread:
                self._countdown_thread.reset()
                self._countdown_thread = None
            self.display.reset()
            self._is_running = False
            self._ev.set()
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            if self._state_watcher:
                self._state_watcher.shutdown()
                self._state_watcher = None

    def recycle(self, refresh_display: bool = True) -> None:
        with self._lock:
            self._holding = self._aborted = False
            self.display[3] = ""
            if refresh_display is True:
                self.update_display(clear=True)

    def close(self) -> None:
        self.reset()

    def _hide(self) -> None:
        with self._lock:
            if self._hidden is False:
                self._hidden = True
                self.display.hide()

    def _show(self) -> None:
        with self._lock:
            if self._hidden is True:
                self._hidden = False
                self.display.show()


class CountdownThread(Thread):
    def __init__(self, status: LaunchStatus, countdown: int = -30) -> None:
        super().__init__(daemon=True, name="Countdown Thread")
        self._status = status
        self._countdown = countdown
        self._hold = False
        self._resume = False
        self._ev = Event()
        self._interval = 1
        self._is_running = True
        self.start()

    @property
    def countdown(self) -> int:
        return self._countdown

    @property
    def is_hold(self) -> bool:
        return self._hold

    def hold(self) -> None:
        if self._countdown and self._countdown < 0 and self._hold is False and self._resume is False:
            self._hold = True
            self._resume = False
            self._interval = None
            self._ev.set()

    def resume(self) -> None:
        if self._countdown and self._countdown < 0 and self._hold is True and self._resume is False:
            self._hold = False
            self._resume = True
            self._interval = 1
            self._ev.set()

    def reset(self) -> None:
        self._is_running = False
        if self.is_alive():
            self._hold = self._resume = False
            self._ev.set()
            self.join()

    def run(self) -> None:
        while self._is_running:
            while not self._ev.wait(self._interval):
                self._countdown += 1
                self._status.countdown = self._countdown
            if self.is_hold is True:
                self._ev.clear()
                continue
            if self._resume is True:
                self._ev.clear()
                self._hold = self._resume = False
                continue



#!/usr/bin/env python3
#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import logging
from argparse import ArgumentParser
from typing import List

from . import CliBase
from ..pdi.asc2_req import Asc2Req
from ..pdi.constants import PdiCommand, Asc2Action
from ..protocol.command_base import CommandBase
from ..protocol.constants import CommandScope
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class Asc2Cmd(CommandBase):
    def __init__(self, tmcc_id: int, req: Asc2Req, server: str = None) -> None:
        if tmcc_id < 1 or tmcc_id > 99:
            raise ValueError("Asc2 ID must be between 1 and 99")
        super().__init__(None, req, tmcc_id, CommandScope.ASC2, server=server)
        self._command = self._build_command()

    def _build_command(self) -> bytes | None:
        return self.command_req.as_bytes

    def _command_prefix(self) -> bytes | None:
        pass

    def _encode_address(self, command_op: int) -> bytes | None:
        pass


class Asc2Cli(CliBase):
    """
    Issue Asc2 Commands.
    """

    @classmethod
    def command_parser(cls) -> ArgumentParser:
        from . import PROGRAM_NAME

        asc2_parser = PyTrainArgumentParser(add_help=False)
        asc2_parser.add_argument("asc2", metavar="Asc2 TMCC ID", type=int, help="Asc2 to fire")

        group = asc2_parser.add_mutually_exclusive_group()
        group.add_argument(
            "-on",
            action="store_true",
            help="Turn Asc2 on",
        )
        group.add_argument(
            "-off",
            action="store_true",
            help="Turn Asc2 off",
        )
        group.add_argument(
            "-hold",
            nargs="?",
            type=float,
            const=1.0,
            help="Turn Asc2 on for specified time",
        )

        asc2_parser.add_argument(
            "-server",
            action="store",
            help=f"IP Address of {PROGRAM_NAME} server, if client. Server communicates with Base 3/LCS SER2",
        )
        # fire command
        return PyTrainArgumentParser(
            "Operate specified Asc2 (1 - 99)",
            parents=[asc2_parser, cls.multi_parser()],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        self._asc2 = self._args.asc2
        self._time = self._args.hold if self._args.hold else 0.0
        self._state = 0 if self._args.off is True else 1
        # adjust time and duration parameters
        if self._state == 1:
            if self._time > 2.55 or (self._args.duration and self._args.duration > 2.55):
                # pik bigger number
                time = self._time if self._time else 0
                duration = self._args.duration if self._args.duration else 0
                self._args.duration = max(time, duration)
                self._time = 0.250 if self.is_server else 0.600
                self._args.duration -= self._time
                self._args.duration = self._args.duration if self._args.duration > 0.0 else 0.0
            elif self._args.duration and self._args.duration <= 2.55:
                self._time = self._args.duration
                self._args.duration = 0
            elif self._time:
                self._args.duration = 0
            else:
                self._time = self._args.duration = 0.0
        else:
            self._time = self._args.duration = 0.0
        req = Asc2Req(self._asc2, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=self._state, time=self._time)
        try:
            cmd = Asc2Cmd(self._asc2, req, server=self._server)
            if self.do_fire:
                cmd.fire(server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from gpiozero import Button

from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum
from .gpio_device import GpioDevice, P


class SystemAdmin(GpioDevice):
    def __init__(self):
        super().__init__()

    def shutdown(
        self,
        shutdown_pin: P,
        hold_time: float = 5,
    ) -> Button:
        """
        Send the system shutdown command to all nodes
        """
        cmd, shutdown_btn, led = self.make_button(shutdown_pin, TMCC1SyncCommandEnum.SHUTDOWN, hold_time=hold_time)
        shutdown_btn.when_held = cmd.as_action()
        return shutdown_btn

    def reboot(
        self,
        reboot_pin: P,
        hold_time: float = 5,
    ) -> Button:
        """
        Send the system restart command to all nodes
        """
        cmd, reboot_btn, led = self.make_button(reboot_pin, TMCC1SyncCommandEnum.REBOOT, hold_time=hold_time)
        reboot_btn.when_held = cmd.as_action()
        return reboot_btn

    def restart(
        self,
        restart_pin: P,
        hold_time: float = 5,
    ) -> Button:
        """
        Send the system restart command to all nodes
        """
        cmd, restart_btn, led = self.make_button(restart_pin, TMCC1SyncCommandEnum.RESTART, hold_time=hold_time)
        restart_btn.when_held = cmd.as_action()
        return restart_btn

    def update(
        self,
        update_pin: P,
        hold_time: float = 5,
    ) -> Button:
        """
        Send the system update command to all nodes
        """
        cmd, update_btn, led = self.make_button(update_pin, TMCC1SyncCommandEnum.UPDATE, hold_time=hold_time)
        update_btn.when_held = cmd.as_action()
        return update_btn

    def upgrade(
        self,
        upgrade_pin: P,
        hold_time: float = 5,
    ) -> Button:
        """
        Send the system upgrade command to all nodes
        """
        cmd, upgrade_btn, led = self.make_button(upgrade_pin, TMCC1SyncCommandEnum.UPGRADE, hold_time=hold_time)
        upgrade_btn.when_held = cmd.as_action()
        return upgrade_btn

    def resync(
        self,
        resync_pin: P,
        hold_time: float = None,
    ) -> Button:
        """
        Send the resync state command to server
        """
        cmd, resync_btn, led = self.make_button(resync_pin, TMCC1SyncCommandEnum.RESYNC, hold_time=hold_time)
        if hold_time:
            resync_btn.when_held = cmd.as_action()
        else:
            resync_btn.when_pressed = cmd.as_action()
        return resync_btn



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

#
import logging
from argparse import ArgumentParser
from typing import List

from . import CliBaseTMCC
from ..protocol.constants import CommandSyntax
from ..protocol.tmcc1.halt_cmd import HaltCmd as HaltCmdTMCC1
from ..protocol.tmcc2.halt_cmd import HaltCmd as HaltCmdTMCC2
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class HaltCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls):
        return PyTrainArgumentParser(
            "Emergency halt; stop all engines and trains",
            parents=[cls.train_parser(), cls.command_format_parser(CommandSyntax.TMCC), cls.cli_parser()],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        try:
            if self.is_train_command or self.is_tmcc2:
                cmd = HaltCmdTMCC2(baudrate=self._baudrate, port=self._port, server=self._server)
            else:
                cmd = HaltCmdTMCC1(baudrate=self._baudrate, port=self._port, server=self._server)
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
import socket
from collections import defaultdict, deque
from queue import Queue
from threading import Thread, Condition, RLock
from time import sleep
from typing import Generic, List, Protocol, Tuple, TypeVar, runtime_checkable, cast

from ..db.component_state import ComponentState
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import TMCC_FIRST_BYTE_TO_INTERPRETER, CommandReq
from ..protocol.constants import (
    BROADCAST_TOPIC,
    DEFAULT_BAUDRATE,
    DEFAULT_PORT,
    DEFAULT_QUEUE_SIZE,
    DEFAULT_VALID_BAUDRATES,
    CommandScope,
    DEFAULT_SERVER_PORT,
    PROGRAM_NAME,
)
from ..protocol.multibyte.multibyte_constants import TMCC2_VARIABLE_INDEX
from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum, SyncCommandDef
from ..protocol.tmcc2.tmcc2_constants import LEGACY_MULTIBYTE_COMMAND_PREFIX
from ..utils.ip_tools import get_ip_address

log = logging.getLogger(__name__)

Message = TypeVar("Message")
Topic = TypeVar("Topic")

SYNCING = CommandReq(TMCC1SyncCommandEnum.SYNCHRONIZING)
SYNC_COMPLETE = CommandReq(TMCC1SyncCommandEnum.SYNCHRONIZED)


class CommandListener(Thread):
    _instance: None = None
    _lock = RLock()

    @classmethod
    def build(
        cls,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        ser2_receiver: bool = True,
        base3_receiver: bool = False,
        server_port: int = None,
    ) -> CommandListener:
        """
        Factory method to create a CommandListener instance
        """
        return CommandListener(
            baudrate=baudrate,
            port=port,
            queue_size=queue_size,
            ser2_receiver=ser2_receiver,
            base3_receiver=base3_receiver,
            server_port=server_port,
        )

    @classmethod
    def get(cls) -> CommandListener:
        if cls._instance is None:
            cls.build()
        return cls._instance

    @classmethod
    def listen_for(
        cls, listener: Subscriber, channel: Topic, address: int = None, command: CommandDefEnum = None, data: int = None
    ):
        cls.build().subscribe(listener, channel, address, command, data)

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def is_running(cls) -> bool:
        # noinspection PyProtectedMember
        return cls._instance is not None and cls._instance._is_running is True

    @classmethod
    def stop(cls) -> None:
        with cls._lock:
            if cls._instance:
                cls._instance.shutdown()

    def __init__(
        self,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        ser2_receiver: bool = True,
        base3_receiver: bool = False,
        server_port: int = None,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        if baudrate not in DEFAULT_VALID_BAUDRATES:
            raise ValueError(f"Invalid baudrate: {baudrate}")
        self._baudrate = baudrate
        self._port = port
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Command Listener")

        # prep our consumer(s)
        self._cv = Condition()
        self._deque = deque(maxlen=DEFAULT_QUEUE_SIZE)
        self._is_running = True
        self._dispatcher = CommandDispatcher.build(queue_size, ser2_receiver, base3_receiver, server_port)

        # get initial state from Base 3 and LCS modules
        self.sync_state()

        # start listener thread
        self.start()

        # prep our producer
        if ser2_receiver:
            from .serial_reader import SerialReader

            self._serial_reader = SerialReader(baudrate, port, self)
        else:
            self._serial_reader = None
        self._ser2_receiver = ser2_receiver
        self._base3_receiver = base3_receiver
        self._filter_updates = base3_receiver is True and ser2_receiver is True

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if CommandListener._instance is None:
                # noinspection PyTypeChecker
                CommandListener._instance = super(CommandListener, cls).__new__(cls)
                CommandListener._instance._initialized = False
            return CommandListener._instance

    def sync_state(self) -> None:
        """
        Get initial system state from the Base 3/LCS Modules
        """
        pass

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @property
    def port(self) -> str:
        return self._port

    def run(self) -> None:
        is_tmcc4 = False
        while self._is_running:
            # process bytes, as long as there are any
            with self._cv:
                if not self._deque:
                    self._cv.wait()  # wait to be notified
            # check if the first bite is in the list of allowable command prefixes
            dq_len = len(self._deque)
            if dq_len >= 3 and int(self._deque[0]) in TMCC_FIRST_BYTE_TO_INTERPRETER:
                # At this point, we have some sort of command. It could be a TMCC1 or TMCC2
                # 3-byte command, or, if there are more than 3 bytes, and the 4th byte is
                # 0xf8 or 0xf9 AND the 5th byte is 0xfb, it could be a 9-byte param command.
                # Try for the 9+ byters first
                cmd_bytes = bytes()
                if (
                    dq_len >= 9
                    and self._deque[3] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                    and self._deque[6] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                ) or (
                    dq_len >= 21
                    and self._deque[7] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                    and self._deque[14] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                ):
                    # if dq_len > 9 and byte 3 is the Variable Command marker, go for more
                    if dq_len >= 9 and self._deque[2] == TMCC2_VARIABLE_INDEX:
                        # byte 5 contains the data word count
                        data_words = int(self._deque[5])
                        command_bytes = (5 + data_words) * 3
                        if dq_len < command_bytes:
                            continue  # wait for more
                        else:
                            last_byte = command_bytes
                    else:  # 4-digit engine/train
                        if (
                            dq_len >= 9
                            and self._deque[3] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                            and self._deque[6] == LEGACY_MULTIBYTE_COMMAND_PREFIX
                        ):
                            last_byte = 9
                        else:
                            last_byte = 21
                            is_tmcc4 = True
                    for _ in range(last_byte):
                        cmd_bytes += self._deque.popleft().to_bytes(1, byteorder="big")
                elif dq_len >= 4 and self._deque[3] == LEGACY_MULTIBYTE_COMMAND_PREFIX:
                    # we could be in the middle of receiving a parameter command, wait a bit longer
                    continue
                else:
                    # assume a 3-byte command
                    for _ in range(3):
                        cmd_bytes += self._deque.popleft().to_bytes(1, byteorder="big")
                    # check for 4-digit addressing
                    dq_len -= 3
                    if dq_len >= 4 and cmd_bytes[1] in {0x00, 0x01}:
                        for _ in range(4):
                            cmd_bytes += self._deque.popleft().to_bytes(1, byteorder="big")
                        dq_len -= 4
                        is_tmcc4 = True
                if cmd_bytes:
                    try:
                        # build_req a CommandReq from the received bytes and send it to the dispatcher
                        self._dispatcher.offer(CommandReq.from_bytes(cmd_bytes, is_tmcc4=is_tmcc4))
                        is_tmcc4 = False
                    except ValueError as ve:
                        log.exception(ve)
            elif dq_len < 3:
                continue  # wait for more bytes
            else:
                # pop this byte and continue; we either received unparsable input
                # or started receiving data mid-command
                log.warning(f"Ignoring {hex(self._deque.popleft())} (deque: {len(self._deque)} bytes)")
        # shut down the dispatcher
        if self._dispatcher:
            self._dispatcher.shutdown()

    def offer(self, data: bytes) -> None:
        from .enqueue_proxy_requests import SYNC_BEGIN_RESPONSE, SYNC_COMPLETE_RESPONSE

        if data:
            with self._cv:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug(f"TMCC CommandListener offered: {data.hex(' ')}")
                # check if these are sync start or end commands
                if data in {SYNC_BEGIN_RESPONSE, SYNC_COMPLETE_RESPONSE}:
                    if data == SYNC_BEGIN_RESPONSE:
                        self._dispatcher.offer(SYNCING)
                    else:
                        self._dispatcher.offer(SYNC_COMPLETE)
                else:
                    self._deque.extend(data)
                    self._cv.notify()

    def shutdown(self) -> None:
        # if specified baudrate was invalid, instance won't have most attributes
        if hasattr(self, "_cv"):
            with self._cv:
                self._is_running = False
                self._cv.notify()
        if hasattr(self, "_serial_reader"):
            if self._serial_reader:
                self._serial_reader.shutdown()
        if hasattr(self, "_dispatcher"):
            if self._dispatcher:
                self._dispatcher.shutdown()
        CommandListener._instance = None

    def subscribe(
        self,
        listener: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> None:
        self._dispatcher.subscribe(listener, channel, address, command, data)

    def unsubscribe(
        self,
        listener: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> None:
        self._dispatcher.unsubscribe(listener, channel, address, command, data)

    def subscribe_any(self, subscriber: Subscriber) -> None:
        self._dispatcher.subscribe_any(subscriber)

    def unsubscribe_any(self, subscriber: Subscriber) -> None:
        self._dispatcher.unsubscribe_any(subscriber)


@runtime_checkable
class Subscriber(Protocol):
    """
    Protocol that all listener callbacks must implement
    """

    def __call__(self, message: Message) -> None: ...


class Channel(Generic[Topic]):
    """
    Part of the publish/subscribe pattern described here:
    https://arjancodes.com/blog/publish-subscribe-pattern-in-python/
    In our case, the "channels" are the valid CommandScopes, a tuple
    consisting of a CommandScope and an TMCC ID/Address, and a
    special "BROADCAST" channel that receives all received commands.
    """

    def __init__(self) -> None:
        self.subscribers: set[Subscriber] = set[Subscriber]()

    def subscribe(self, subscriber: Subscriber[Message]) -> None:
        self.subscribers.add(subscriber)

    def unsubscribe(self, subscriber: Subscriber[Message]) -> None:
        self.subscribers.remove(subscriber)

    def publish(self, message: Message) -> None:
        for subscriber in self.subscribers:
            try:
                subscriber(message)
            except Exception as e:
                log.warning(f"CommandDispatcher: Error publishing {message}; see log for details")
                log.exception(e)


class CommandDispatcher(Thread, Generic[Topic, Message]):
    """
    The CommandDispatcher thread receives parsed CommandReqs from the
    CommandListener and dispatches them to subscribing listeners
    """

    _instance = None
    _lock = RLock()

    @classmethod
    def build(
        cls,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        ser2_receiver: bool = False,
        base3_receiver: bool = False,
        server_port: int = None,
    ) -> CommandDispatcher:
        """
        Factory method to create a CommandDispatcher instance
        """
        return CommandDispatcher(queue_size, ser2_receiver, base3_receiver, server_port)

    @classmethod
    def get(cls) -> CommandDispatcher:
        if cls._instance is None:
            raise AttributeError("Command Dispatcher not yet created")
        return cls._instance

    @classmethod
    def listen_for(
        cls,
        listener: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ):
        cls.build().subscribe(listener, channel, address, command, data)

    @classmethod
    def is_running(cls) -> bool:
        # noinspection PyProtectedMember
        return cls._instance is not None and cls._instance._is_running is True

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if CommandDispatcher._instance is None:
                CommandDispatcher._instance = super(CommandDispatcher, cls).__new__(cls)
                CommandDispatcher._instance._initialized = False
            return CommandDispatcher._instance

    def __init__(
        self,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        ser2_receiver: bool = False,
        base3_receiver: bool = False,
        server_port: int = None,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        from .enqueue_proxy_requests import EnqueueProxyRequests

        super().__init__(daemon=True, name=f"{PROGRAM_NAME} TMCC Command Dispatcher")
        self._is_ser2_receiver = ser2_receiver
        self._is_base3_receiver = base3_receiver
        self._filter_updates = base3_receiver is True and ser2_receiver is True
        self._channels: dict[Topic | Tuple[Topic, int], Channel[Message]] = defaultdict(Channel)
        self._cv = Condition()  # for queue access
        self._chanel_lock = Condition()  # for publishing
        self._client_lock = Condition()  # for updating client state
        self._is_running = True
        self._queue = Queue[CommandReq](queue_size)
        self._broadcasts = False
        self._is_server = ser2_receiver is True or base3_receiver is True
        if server_port:
            self._server_port = server_port
        elif EnqueueProxyRequests.is_built():
            self._server_port = EnqueueProxyRequests.server_port()
        elif self._is_server is True:
            self._server_port = DEFAULT_SERVER_PORT
            # TODO: this is really an error, fix tests accordingly
            # raise AttributeError("EnqueueProxyRequests not yet built")
        else:
            self._server_port = None
        self._server_ips = get_ip_address()
        self.start()

    def run(self) -> None:
        while self._is_running:
            with self._cv:
                if self._queue.empty():
                    self._cv.wait()
            if self._queue.empty():  # we need to do a second check in the event we're being shutdown
                continue
            cmd = self._queue.get()
            try:
                # publish dispatched commands to listeners on the command scope,
                if isinstance(cmd, CommandReq):
                    # if command is a TMCC1 Halt, send to everyone
                    if cmd.is_halt:
                        if self._filter_updates is True and cmd.is_filtered is True:
                            log.debug(f"Filtering client update: {cmd}")
                        else:
                            self.publish_all(cmd)
                    # if command is a legacy-style halt, just send to engines and trains
                    elif cmd.is_system_halt:
                        self.publish_all(cmd, [CommandScope.ENGINE, CommandScope.TRAIN])
                    # otherwise, send to the interested parties
                    else:
                        if cmd.is_data is True:
                            self.publish((cmd.scope, cmd.address, cmd.command, cmd.data), cmd)
                        self.publish((cmd.scope, cmd.address, cmd.command), cmd)
                        self.publish((cmd.scope, cmd.address), cmd)
                        self.publish(cmd.scope, cmd)
                    if self._broadcasts:
                        self.publish(BROADCAST_TOPIC, cmd)
                    # update state on all clients
                    if self._server_port is not None:
                        """
                        When we are listening to broadcasts from both the Base 3 and a Ser2, the
                        TMCC commands broadcast from the Base 3 are also sent out via the Ser2.
                        However, the command stream on the Ser2 lags the base. This can result in
                        command conflicts. For example, when sending speed change commands to the
                        Base 3 to go to speed 10, then 20, then 30, then 40, these directives
                        are echoed back almost instantly via the Base 3 and used to update the
                        LCD display. But, when the Ser2 echos the same commands a few seconds later,
                        the display can look like 10, 20, 30, 10, 40, 30, 40, because the received
                        commands from the Ser2 arrive out of sync.

                        The fix is to filter out the propagation of the Ser2 commands if an equivalent
                        command is broadcast from the Base 3, and if we are listening to both the
                        Base 3 and Ser2. Only a subset of the TMCC commands are sent out via the Base 3.
                        These commands are all marked as "filtered", and are excluded here as well as
                        in ComponentStateStore.
                        """
                        if self._filter_updates is True and cmd.is_filtered is True:
                            log.debug(f"Filtering client update: {cmd}")
                        else:
                            self.update_client_state(cmd)
            except Exception as e:
                log.warning(f"CommandDispatcher: Error publishing {cmd}; see log for details")
                log.exception(e)
            finally:
                self._queue.task_done()

    @property
    def is_ser2_receiver(self) -> bool:
        """
        Returns whether the instance is a Ser2 receiver or not.

        The SER2 receiver is a specific type of receiver characterized by the
        presence of an LCS Ser2 that is broadcasting TMCC commands.

        Returns
        -------
        bool
            True if the instance has the SER2 receiver capability, else False.
        """
        return self._is_ser2_receiver

    @property
    def is_base3_receiver(self) -> bool:
        return self._is_base3_receiver

    @property
    def is_filter_updates(self) -> bool:
        return self._filter_updates

    def signal_clients(
        self,
        option: CommandReq | TMCC1SyncCommandEnum = TMCC1SyncCommandEnum.QUIT,
        client: str = None,
        port: int = None,
    ) -> None:
        if isinstance(option, TMCC1SyncCommandEnum):
            option = CommandReq(option)
        self.update_client_state(option, client=client, port=port)

    def signal_clients_on(
        self, option: CommandReq | TMCC1SyncCommandEnum = TMCC1SyncCommandEnum.QUIT, client: str = None
    ) -> None:
        from .enqueue_proxy_requests import EnqueueProxyRequests

        if isinstance(option, TMCC1SyncCommandEnum):
            option = CommandReq(option)
        for client_ip, port in EnqueueProxyRequests.clients():
            if client_ip == client:
                node_scope = cast(SyncCommandDef, option.command_def).is_node_scope
                self.update_client_state(option, client=client, port=port)
                if node_scope is True:
                    return

    def signal_client(
        self,
        option: CommandReq | TMCC1SyncCommandEnum = TMCC1SyncCommandEnum.QUIT,
        client: str = None,
        port: int = None,
    ) -> None:
        if isinstance(option, TMCC1SyncCommandEnum):
            option = CommandReq(option)
        self.update_client_state(option, client=client, port=port)

    # noinspection DuplicatedCode
    def update_client_state(self, command: CommandReq, client: str = None, port: int = None):
        """
        Update all PyTrain clients with the dispatched command. Used to keep
        client states in sync with server
        """
        from .enqueue_proxy_requests import EnqueueProxyRequests

        if client is None:
            clients = EnqueueProxyRequests.clients()
        else:
            if port is None:
                port = DEFAULT_SERVER_PORT
            clients = {(client, port)}
        # noinspection PyTypeChecker
        for client, port in clients:
            if client in self._server_ips and port == self._server_port:
                print(f"Skipping update of {client}:{port} {command}")
                continue
            try:
                with self._client_lock:
                    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                        s.connect((client, port))
                        s.sendall(command.as_bytes)
                        _ = s.recv(32)
                    self._client_lock.notify_all()
            except ConnectionRefusedError:
                # ignore disconnects; client will receive state update on reconnect
                pass
            except Exception as e:
                log.warning(f"Exception while sending TMCC state update {command} to {client}")
                log.exception(e)

    # noinspection PyTypeChecker
    def send_current_state(self, client_ip: str, client_port: int = None):
        """
        When a new client attaches to the server, immediately send it all know
        component states. They will be updated as needed (see update_client_state).
        """
        client_port = client_port if client_port else self._server_port
        if client_port is not None:
            from ..db.component_state_store import ComponentStateStore
            from .enqueue_proxy_requests import EnqueueProxyRequests

            # send starting state sync message
            self.send_state_packet(client_ip, client_port, EnqueueProxyRequests.sync_begin_response())
            store = ComponentStateStore.get()
            for scope in store.scopes():
                if scope == CommandScope.SYNC:
                    continue
                for address in store.addresses(scope):
                    with self._client_lock:
                        state: ComponentState = store.query(scope, address)
                        if state is not None:
                            state_bytes = state.as_bytes()
                            if isinstance(state_bytes, bytes):
                                state_bytes = [state_bytes]
                            elif isinstance(state_bytes, list):
                                pass
                            else:
                                raise TypeError(f"Invalid state type: {type(state_bytes)}")
                            do_pause = False
                            for state_packet in state_bytes:
                                if not state_packet:
                                    continue
                                if do_pause is True:
                                    sleep(0.01)
                                try:
                                    self.send_state_packet(client_ip, client_port, state_packet)
                                    do_pause = True
                                except Exception as e:
                                    log.warning(f"Exception sending state update {state} to {client_ip}:{client_port}")
                                    log.exception(e)
                        self._client_lock.notify_all()
            # send sync complete message
            self.send_state_packet(client_ip, client_port, EnqueueProxyRequests.sync_complete_response())

    def send_state_packet(self, client_ip: str, client_port: int, state: ComponentState | bytes):
        client_port = client_port if client_port else self._server_port
        packet: bytes | None = None
        try:
            if isinstance(state, bytes):
                packet = state
            elif isinstance(state, ComponentState):
                # TODO: for safety, make this code handle if bytes are returned as list
                packet = state.as_bytes()
            if isinstance(packet, list):
                log.error(f"send_state_packet: Invalid state type: {type(packet)}")

            if packet:  # we can only send states for tracked conditions
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.connect((client_ip, client_port))
                    s.sendall(packet)
                    _ = s.recv(32)
        except Exception as e:
            log.warning(f"Exception sending TMCC state update {state} to {client_ip}:{client_port}")
            log.exception(e)

    @property
    def broadcasts_enabled(self) -> bool:
        return self._broadcasts

    def offer(self, cmd: CommandReq, from_pdi: bool = False) -> None:
        """
        Receive a command from the TMCC listener thread and dispatch it to subscribers.
        We do this in a separate thread so that the listener thread doesn't fall behind.
        """
        if isinstance(cmd, CommandReq):
            with self._cv:
                self._queue.put(cmd)
                self._cv.notify_all()  # wake up receiving thread
                if from_pdi is True:
                    pass  # TODO: prevent receiving same command from ser2 stream

    def shutdown(self) -> None:
        with self._cv:
            self._is_running = False
            self._cv.notify_all()
        CommandDispatcher._instance = None

    @staticmethod
    def _make_channel(
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> CommandScope | Tuple:
        if channel is None:
            raise ValueError("Channel required")
        elif address is None:
            return channel
        elif command is None:
            return channel, address
        elif data is None:
            return channel, address, command
        elif command.command_def.is_data and data is not None:
            return channel, address, command, data
        else:
            raise TypeError("Command must be Topic or CommandReq")

    def publish_all(self, message: Message, channels: List[CommandScope] = None) -> None:
        with self._chanel_lock:
            if channels is None:  # send to everyone!
                for channel in self._channels:
                    self._channels[channel].publish(message)
            else:
                # send only to select channels and tuples with that channel
                for channel in self._channels.keys():
                    if channel in channels or (isinstance(channel, tuple) and channel[0] in channels):
                        self._channels[channel].publish(message)

    def publish(self, channel: Topic, message: Message) -> None:
        with self._chanel_lock:
            if channel in self._channels:  # otherwise, we would create a channel simply by referencing i
                try:
                    self._channels[channel].publish(message)
                except Exception as e:
                    log.exception(f"Exception publishing {message} to {channel}", exc_info=e)
            self._chanel_lock.notify_all()

    def subscribe(
        self,
        subscriber: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> None:
        with self._chanel_lock:
            if channel == BROADCAST_TOPIC:
                self.subscribe_any(subscriber)
            else:
                self._channels[self._make_channel(channel, address, command, data)].subscribe(subscriber)

    def unsubscribe(
        self,
        subscriber: Subscriber,
        channel: Topic,
        address: int = None,
        command: CommandDefEnum = None,
        data: int = None,
    ) -> None:
        with self._chanel_lock:
            if channel == BROADCAST_TOPIC:
                self.unsubscribe_any(subscriber)
            else:
                channel = self._make_channel(channel, address, command, data)
                if channel in self._channels:
                    self._channels[channel].unsubscribe(subscriber)
                    if len(self._channels[channel].subscribers) == 0:
                        del self._channels[channel]

    def subscribe_any(self, subscriber: Subscriber) -> None:
        # receive broadcasts
        with self._chanel_lock:
            self._channels[BROADCAST_TOPIC].subscribe(subscriber)
            self._broadcasts = True

    def unsubscribe_any(self, subscriber: Subscriber) -> None:
        # receive broadcasts
        with self._chanel_lock:
            self._channels[BROADCAST_TOPIC].unsubscribe(subscriber)
            if not self._channels[BROADCAST_TOPIC].subscribers:
                self._broadcasts = False



from .tmcc1_command import TMCC1Command
from .tmcc1_constants import TMCC1HaltCommandEnum

from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT, DEFAULT_ADDRESS


class HaltCmd(TMCC1Command):
    """
    Issue TMCC1 Halt command; stops all engines in motion
    """

    def __init__(self, baudrate: int = DEFAULT_BAUDRATE, port: str = DEFAULT_PORT, server: str = None) -> None:
        req = CommandReq(TMCC1HaltCommandEnum.HALT)
        super().__init__(TMCC1HaltCommandEnum.HALT, req, DEFAULT_ADDRESS, 0, None, baudrate, port, server)
        self._command = self._build_command()



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from __future__ import annotations

import logging

from gpiozero import Button

from ..db.component_state import SwitchState
from ..db.engine_state import EngineState, TrainState
from ..db.irda_state import IrdaState
from ..db.block_state import BlockState
from ..db.component_state_store import ComponentStateStore
from ..db.state_watcher import StateWatcher
from ..gpio.gpio_handler import GpioHandler, P
from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope, Direction
from ..protocol.multibyte.multibyte_constants import TMCC2RailSoundsDialogControl
from ..protocol.tmcc1.tmcc1_constants import TMCC1_RESTRICTED_SPEED, TMCC1EngineCommandEnum
from ..protocol.tmcc2.tmcc2_constants import TMCC2_RESTRICTED_SPEED, TMCC2EngineCommandEnum
from ..utils.validations import Validations

log = logging.getLogger(__name__)


class Block:
    """
    This class represents a Block, a section of a railway network, and manages its state and status.

    The Block is designed to monitor presence and state information for engines or trains traversing through it.
    It integrates with sensors and other components to track occupancy, direction, and events like entering
    or exiting a portion of the Block. Blocks can also be linked together to form a sequence in the network,
    with support for directionality and connections to adjacent Blocks. Furthermore, Blocks interact with switches
    to handle diverging or converging paths and broadcast their state to external systems.

    Args:
        block_id (int): Unique identifier for the block.
        block_name (Optional[str]): Name of the block, if any.
        sensor_track_id (Optional[int]): Unique identifier for the IRDA sensor used to track trains in the block.
        enter_pin (Optional[P]): Pin for the entry button used to mark the block as occupied.
        slow_pin (Optional[P]): Pin for the slow section button in the block.
        stop_pin (Optional[P]): Pin for the stop section button in the block.
        left_to_right (bool): Default directionality of the block (left to right if True).
        dialog (bool): Determines whether dialogs are triggered for block events (default is True).

    Errors:
        AttributeError: Raised in situations such as invalid block id, missing switches, or incorrect configuration.
    """

    @classmethod
    def button(cls, pin: P) -> Button:
        return GpioHandler.make_button(pin)

    # noinspection PyTypeChecker
    def __init__(
        self,
        block_id: int,
        block_name: str = None,
        sensor_track_id: int = None,
        enter_pin: P = None,
        slow_pin: P = None,
        stop_pin: P = None,
        left_to_right: bool = True,
        dialog: bool = True,
    ) -> None:
        # check if block_id is valid
        Validations.validate_int(
            block_id,
            min_value=1,
            max_value=99,
            label="Block ID",
        )
        Validations.validate_int(
            sensor_track_id,
            min_value=1,
            max_value=99,
            label="Sensor Track TMCC ID",
            allow_none=True,
        )
        if ComponentStateStore.get_state(CommandScope.BLOCK, block_id, create=False):
            raise AttributeError(f"Block ID {block_id} is in use")
        self._block_id = block_id
        self._block_name = block_name
        if sensor_track_id:
            self._sensor_track: IrdaState = ComponentStateStore.get_state(CommandScope.IRDA, sensor_track_id)
        else:
            self._sensor_track = None

        self._enter_btn = self.button(enter_pin) if enter_pin else None
        self._slow_btn = self.button(slow_pin) if slow_pin else None
        self._stop_btn = self.button(stop_pin) if stop_pin else None

        self._prev_block: Block | None = None
        self._next_block: Block | None = None
        self._current_motive: EngineState | TrainState | None = None
        self._motive_direction = None
        self._original_speed: int | None = None
        self._switch: SwitchState = None
        self._thru_block: Block = None
        self._out_block: Block = None
        self._left_to_right = left_to_right if left_to_right is not None else True
        self._dialog = dialog
        self._order_activated = []
        self._order_deactivated = []

        # add handlers for state change
        if self._enter_btn:
            self._enter_btn.when_activated = self.signal_occupied_enter
            self._enter_btn.when_activated = self.signal_occupied_exit
        if self._slow_btn:
            self._slow_btn.when_activated = self.signal_slow_enter
            self._slow_btn.when_deactivated = self.signal_slow_exit
        if self._stop_btn:
            self._stop_btn.when_activated = self.signal_stop_enter
            self._stop_btn.when_deactivated = self.signal_stop_exit

        # start thread if sensor track specified, we also delay calling super until
        # buttons have been created
        self._sensor_track_watcher = self._switch_watcher = None
        if self.sensor_track:
            self._sensor_track_watcher = StateWatcher(self.sensor_track, self._cache_motive)
        # finally, update corresponding state record on all nodes
        self.broadcast_state()
        self._block_state = ComponentStateStore.get_state(CommandScope.BLOCK, self.block_id)

    def __repr__(self) -> str:
        nm = f" {self.name}" if self.name else ""
        oc = f" Occupied: {self.is_occupied if self.is_occupied is not None else 'Unknown'}"
        dr = f" Direction: {self.direction.name}" if self.direction else ""
        return f"Block{nm} #{self.block_id}{oc}{dr}"

    @property
    def name(self) -> str:
        return self._block_name

    @property
    def block_id(self) -> int:
        return self._block_id

    @property
    def sensor_track(self) -> IrdaState:
        """
        Return the IrdaState at the end of the block, if any.
        Required to control the specific engine/train in the block
        """
        return self._sensor_track

    @property
    def switch(self) -> SwitchState:
        """
        Return the SwitchState at the end of the block, if any
        """
        return self._switch

    @property
    def occupied_by(self) -> EngineState | TrainState | None:
        """
        Return the EngineState or TrainState of the entity
        currently traversing or stopped in the block
        """
        return self._current_motive

    @property
    def occupied_direction(self) -> Direction:
        """
        What direction is the train in the block moving,
        L to R or R to L.
        """
        return self._motive_direction

    @property
    def is_occupied(self) -> bool:
        """
        Is the block occupied?
        """
        return (
            (self.occupied_by and self.occupied_direction == self.direction)
            or self.is_entered is True
            or self.is_slowed is True
            or self.is_stopped is True
        )

    @property
    def is_clear(self) -> bool:
        """
        Is the block unoccupied?
        """
        return self.is_occupied is False

    @property
    def is_entered(self) -> bool:
        """
        Is the "entered" block portion occupied?
        """
        return self._enter_btn.is_active if self._enter_btn else None

    @property
    def is_slowed(self) -> bool:
        """
        Is the "slow" block portion occupied?
        """
        return self._slow_btn.is_active if self._slow_btn else None

    @property
    def is_stopped(self) -> bool:
        """
        Is the "stopped" block portion occupied?
        """
        return self._stop_btn.is_active if self._stop_btn else None

    @property
    def is_dialog(self) -> bool:
        """
        Returns True if tower/Engineer dialogs are given in
        response to block events in this block
        """
        return self._dialog

    @is_dialog.setter
    def is_dialog(self, dialog: bool) -> None:
        self._dialog = dialog

    @property
    def prev_block(self) -> Block | None:
        return self._prev_block

    @prev_block.setter
    def prev_block(self, block: Block) -> None:
        self._prev_block = block
        if block and block.next_block != self:
            block.next_block = self

    @property
    def next_block(self) -> Block | None:
        return self._next_block

    @next_block.setter
    def next_block(self, block: Block) -> None:
        self._next_block = block
        if block and block.prev_block != self:
            block.prev_block = self

    @property
    def is_left_to_right(self) -> bool:
        return self._left_to_right

    @property
    def is_right_to_left(self) -> bool:
        return not self._left_to_right

    @property
    def direction(self) -> Direction:
        """
        Block directionality. Can be L to R (default) or R to L.
        The IR Sensor track must be at the beginning of the block,
        so for an L to R block, the sensor track is placed to the
        left of the block entrance.
        """
        return Direction.L2R if self.is_left_to_right else Direction.R2L

    @property
    def state(self) -> BlockState:
        """
        Return the BlockState associated with this block
        """
        return self._block_state

    def broadcast_state(self):
        """
        Send current block state to PyTrain server and from
        there to all clients
        """
        from ..comm.comm_buffer import CommBuffer
        from ..pdi.block_req import BlockReq

        block_req = BlockReq(self)
        CommBuffer.get().update_state(block_req)

    def next_switch(self, switch_tmcc_id, thru_block: Block, out_block: Block) -> None:
        Validations.validate_int(
            switch_tmcc_id,
            min_value=1,
            max_value=99,
            label="Switch TMCC ID",
        )
        if thru_block is None or out_block is None:
            raise AttributeError("Thru and Out blocks cannot be None")
        if switch_tmcc_id:
            if thru_block is None or out_block is None:
                raise AttributeError("Thru and Out blocks cannot be None")
            if self == thru_block or self == out_block or thru_block == out_block:
                raise AttributeError("Thru and Out blocks cannot be the same block")
            self._switch: SwitchState = ComponentStateStore.get_state(CommandScope.SWITCH, switch_tmcc_id)
            if self._switch is None:
                raise AttributeError(f"Switch {switch_tmcc_id} not found")
            self._thru_block = thru_block
            self._out_block = out_block
            self._switch_watcher = StateWatcher(self.switch, self.respond_to_thrown_switch)
            # force call to the method that sets next_block
            self.respond_to_thrown_switch()
        else:
            raise AttributeError("Switch TMCC ID cannot be None")

    def signal_occupied_enter(self) -> None:
        log.info(f"Block {self.block_id} occupied enter")
        if 1 not in self._order_activated:
            self._order_activated.append(1)
        self.broadcast_state()

    def signal_occupied_exit(self) -> None:
        log.info(f"Block {self.block_id} occupied exit")
        if 1 not in self._order_deactivated:
            self._order_deactivated.append(1)
        # if we are traversing this block in reverse, which we know
        # because the motive direction differs from the defined block
        # direction, clear the motive info
        if self.occupied_direction and self.occupied_direction != self.direction:
            self._original_speed = None
            self._current_motive = None
            self._motive_direction = None
        self.broadcast_state()

    def signal_slow_enter(self) -> None:
        log.info(f"Block {self.block_id} slow enter")
        if 2 not in self._order_activated:
            self._order_activated.append(2)
        if self.occupied_direction == self.direction:
            if self.next_block and self.next_block.is_occupied:
                self.slow_down()
            elif self.next_block and self.next_block.is_clear:
                self.do_dialog(TMCC2RailSoundsDialogControl.ENGINEER_ALL_CLEAR)
        self.broadcast_state()

    def signal_slow_exit(self) -> None:
        log.info(f"Block {self.block_id} slow exit")
        self.broadcast_state()

    def signal_stop_enter(self) -> None:
        log.info(f"Block {self.block_id} stop enter")
        if 3 not in self._order_activated:
            self._order_activated.append(3)
        # if the next block is occupied, stop the train in this block immediately
        if self.occupied_direction == self.direction:
            if self.next_block and self.next_block.is_occupied:
                self.stop_immediate()
            # do dialog, if enabled
            elif self.next_block and self.next_block.is_clear:
                self.do_dialog(TMCC2RailSoundsDialogControl.ENGINEER_ALL_CLEAR)
        self.broadcast_state()

    def signal_stop_exit(self) -> None:
        log.info(f"Block {self.block_id} stop exit")
        if 3 not in self._order_deactivated:
            self._order_deactivated.append(3)
        # if exit was fired in the correct order, clear the block
        self.clear_block_info()
        if self._prev_block and self.is_clear:
            self._prev_block.next_block_clear(self)
        self.broadcast_state()

    def clear_block_info(self):
        self._original_speed = None
        self._current_motive = None
        self._motive_direction = None
        self._order_activated.clear()
        self._order_deactivated.clear()

    def next_block_clear(self, signaling_block: Block) -> None:
        log.info(f"Block {self.block_id} orig speed: {self._original_speed} Block {signaling_block.block_id} Clear")
        if signaling_block.is_occupied is False:
            self.resume_speed()

    def slow_down(self):
        if self.next_block and self.next_block.is_occupied:
            from ..protocol.sequence.ramped_speed_req import RampedSpeedReq

            if self._current_motive:
                restricted_speed = TMCC2_RESTRICTED_SPEED if self._current_motive.is_legacy else TMCC1_RESTRICTED_SPEED
                self._original_speed = self._current_motive.speed
                if self._original_speed > restricted_speed:
                    self.do_dialog(TMCC2RailSoundsDialogControl.TOWER_SPEED_RESTRICTED)
                    scope = self._current_motive.scope
                    tmcc_id = self._current_motive.tmcc_id
                    is_tmcc = self._current_motive.is_tmcc
                    req = RampedSpeedReq(tmcc_id, "restricted", scope, is_tmcc)
                    req.send()

    def stop_immediate(self):
        if self._current_motive:
            if self._original_speed is None:
                self._original_speed = self._current_motive.speed
            log.info(f"Immediate stop; previous speed: {self._original_speed}")
            self.do_dialog(TMCC2RailSoundsDialogControl.TOWER_SPEED_STOP_HOLD)
            scope = self._current_motive.scope
            tmcc_id = self._current_motive.tmcc_id
            if self._current_motive.is_tmcc is True:
                CommandReq(TMCC1EngineCommandEnum.STOP_IMMEDIATE, tmcc_id, scope=scope).send()
            else:
                CommandReq(TMCC2EngineCommandEnum.STOP_IMMEDIATE, tmcc_id, scope=scope).send()
        elif self.is_occupied is True:
            # send a stop to all engines, as otherwise, we could have a crash
            if self.is_dialog:
                self.do_dialog(5)
            else:
                CommandReq(TMCC1EngineCommandEnum.BLOW_HORN_ONE, 99).send()
            CommandReq(TMCC1EngineCommandEnum.STOP_IMMEDIATE, 99).send()

    def resume_speed(self) -> None:
        from ..protocol.sequence.ramped_speed_req import RampedSpeedReq

        if self._current_motive and self._original_speed:
            scope = self._current_motive.scope
            tmcc_id = self._current_motive.tmcc_id
            is_tmcc = self._current_motive.is_tmcc
            log.info(f"Resume Speed: {self._original_speed} for {scope.title} {tmcc_id}")
            self.do_dialog(TMCC2RailSoundsDialogControl.TOWER_DEPARTURE_GRANTED)
            req = RampedSpeedReq(tmcc_id, self._original_speed, scope, is_tmcc)
            req.send()

    def do_dialog(self, dialog: CommandDefEnum | int) -> None:
        if self.is_dialog and self._current_motive:
            scope = self._current_motive.scope
            tmcc_id = self._current_motive.tmcc_id
            if isinstance(dialog, int):
                req = CommandReq.build(TMCC2EngineCommandEnum.NUMERIC, tmcc_id, data=dialog, scope=scope)
            else:
                log.info(f"Do dialog {dialog.title} for {scope.title} {tmcc_id}")
                req = CommandReq.build(dialog, tmcc_id, scope=scope)
            req.send()

    def _cache_motive(self) -> None:
        scope = "Train" if self.sensor_track.is_train else "Engine"
        if scope == "Train":
            last_id = self.sensor_track.last_train_id
            prod_type = f"{self.sensor_track.product_type} " if self.sensor_track.product_type != "NA" else " Engine "
            eid = f" ({prod_type}{self.sensor_track.last_engine_id})"
        else:
            last_id = self.sensor_track.last_engine_id
            eid = ""
        ld = "L -> R" if self.is_left_to_right else "R -> L"
        log.info(
            f"Cache Motive called, {self.sensor_track.tmcc_id} {scope} {last_id}{eid} "
            f"{ld} {self.sensor_track.last_direction}"
        )
        # we want to record the info on the train in the block whether it
        # is coming or going; although if it is going, we have to figure out
        # how to clear it...
        last_motive = self.occupied_by
        if self.sensor_track.is_train is True and self.sensor_track.last_train_id:
            self._current_motive = ComponentStateStore.get_state(CommandScope.TRAIN, self.sensor_track.last_train_id)
        elif self.sensor_track.is_engine is True and self.sensor_track.last_engine_id:
            self._current_motive = ComponentStateStore.get_state(CommandScope.ENGINE, self.sensor_track.last_engine_id)
        else:
            # a train is passing out of this block
            self._current_motive = None
            self._motive_direction = None

        if self._current_motive:
            self._original_speed = self._current_motive.speed
            self._motive_direction = self.sensor_track.last_direction
        else:
            self._original_speed = None
            self._motive_direction = None
        if last_motive != self.occupied_by:
            self.broadcast_state()

    def respond_to_thrown_switch(self) -> None:
        if self.switch:
            log.info(f"{self.switch} Thru: {self._thru_block} Out: {self._out_block}")
            if self.switch.is_through:
                self.next_block = self._thru_block
                self._out_block.prev_block = None
            elif self.switch.is_out:
                self.next_block = self._out_block
                self._thru_block.prev_block = None
            else:
                return
            if self.next_block.is_occupied is True:
                if self._stop_btn.is_active:
                    self.signal_stop_enter()
                elif self._slow_btn.is_active:
                    self.signal_slow_enter()
            elif self.next_block.is_clear:
                self.next_block_clear(self.next_block)
            self.broadcast_state()



from .tmcc2_command import TMCC2Command
from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT, DEFAULT_ADDRESS
from .tmcc2_constants import TMCC2HaltCommandEnum


class HaltCmd(TMCC2Command):
    """
    Issue TMCC2 Halt command; stops all engines in motion
    """

    def __init__(self, baudrate: int = DEFAULT_BAUDRATE, port: str = DEFAULT_PORT, server: str = None) -> None:
        req = CommandReq(TMCC2HaltCommandEnum.HALT, DEFAULT_ADDRESS)
        super().__init__(TMCC2HaltCommandEnum.HALT, req, DEFAULT_ADDRESS, 0, None, baudrate, port, server)
        self._command = self._build_command()



import time

# noinspection PyPackageRequirements
import pytest

# noinspection PyProtectedMember
from src.pytrain.comm.command_listener import CommandListener, CommandDispatcher
from src.pytrain.db.component_state import SwitchState
from src.pytrain.db.component_state_store import ComponentStateStore
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import CommandScope
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum
from test.test_base import TestBase


@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmpdir) -> None:
    """
    Fixture to execute asserts before and after a test is run
    """
    # Setup: fill with any logic you want

    yield  # this is where the testing happens

    # Teardown : fill with any logic you want
    if CommandListener.is_built():
        CommandListener().shutdown()
    assert CommandListener.is_built() is False

    if CommandDispatcher.is_built():
        CommandDispatcher().shutdown()
    assert CommandDispatcher.is_built() is False


# noinspection PyMethodMayBeStatic
class TestComponentStateStore(TestBase):
    def test_component_state_store_basic(self) -> None:
        # create a dispatcher to serve the store
        dispatcher = CommandDispatcher.build()
        assert dispatcher is not None
        assert dispatcher.is_running()

        # create a store
        store = ComponentStateStore(listeners=(dispatcher,))
        store.listen_for([CommandScope.ENGINE, CommandScope.TRAIN, CommandScope.ACC, CommandScope.SWITCH])
        assert store is not None
        assert store.is_empty

        # add some state
        sw_out = CommandReq.build(TMCC1SwitchCommandEnum.OUT, 15)
        dispatcher.offer(sw_out)
        time.sleep(0.1)
        assert store.is_empty is False
        sw_15_state: SwitchState = store.query(CommandScope.SWITCH, 15)
        assert sw_15_state is not None
        assert sw_15_state.last_updated is not None
        assert sw_15_state.last_command == sw_out
        assert sw_15_state.is_known is True
        assert sw_15_state.state == TMCC1SwitchCommandEnum.OUT
        assert sw_15_state.is_out is True
        assert sw_15_state.is_through is False

        # throw to through
        sw_through = CommandReq.build(TMCC1SwitchCommandEnum.THRU, 15)
        dispatcher.offer(sw_through)
        time.sleep(0.1)
        assert sw_15_state is not None
        assert sw_15_state.last_updated is not None
        assert sw_15_state.last_command == sw_through
        assert sw_15_state.is_known is True
        assert sw_15_state.state == TMCC1SwitchCommandEnum.THRU
        assert sw_15_state.is_out is False
        assert sw_15_state.is_through is True

        # set the address of a different switch. should not cause state to be known
        sw_addr = CommandReq.build(TMCC1SwitchCommandEnum.SET_ADDRESS, 47)
        dispatcher.offer(sw_addr)
        time.sleep(0.1)
        sw_47_state: SwitchState = store.query(CommandScope.SWITCH, 47)
        assert sw_47_state is not None
        assert sw_47_state.last_updated is not None
        assert sw_47_state.last_command == sw_addr
        assert sw_47_state.is_known is False
        assert sw_47_state.state is None
        assert sw_47_state.is_out is False
        assert sw_47_state.is_through is False

        # throw the switch to out, state should now be set
        sw_out.address = 47
        assert sw_out.address == 47
        dispatcher.offer(sw_out)
        time.sleep(0.1)
        sw_47_state: SwitchState = store.query(CommandScope.SWITCH, 47)
        assert sw_47_state is not None
        assert sw_47_state.last_updated is not None
        assert sw_47_state.last_command == sw_out
        assert sw_47_state.is_known is True
        assert sw_47_state.state is TMCC1SwitchCommandEnum.OUT
        assert sw_47_state.is_out is True
        assert sw_47_state.is_through is False



from __future__ import annotations

import abc
import logging
from abc import ABC
from collections import defaultdict
from enum import unique, Enum
from typing import TypeVar, List

from .asc2_req import Asc2Req
from .bpc2_req import Bpc2Req
from .constants import (
    FriendlyMixins,
    PdiCommand,
    Asc2Action,
    Bpc2Action,
    IrdaAction,
    Ser2Action,
    Stm2Action,
    WiFiAction,
    CommonAction,
    PdiAction,
    D4Action,
)
from .irda_req import IrdaReq
from .pdi_req import PdiReq
from .stm2_req import Stm2Req
from ..protocol.constants import Mixins, CommandScope

T = TypeVar("T", bound=PdiReq)

log = logging.getLogger(__name__)


class PdiDeviceConfig(ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(self, device: PdiDevice, cmd: T) -> None:
        self._device = device
        self._tmcc_id: int = cmd.tmcc_id if cmd is not None else 0

    @property
    def device(self) -> PdiDevice:
        return self._device

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @property
    @abc.abstractmethod
    def state_requests(self) -> List[T]: ...


class Acs2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Asc2Req) -> None:
        super().__init__(PdiDevice.ASC2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # Acc mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL1))
        elif self._mode == 1:
            # Acc mode, 1 TMCC ID, latching
            cmds.append(Asc2Req(self.tmcc_id, action=Asc2Action.CONTROL2))
        elif self._mode == 2:
            # Switch mode, pulsed, 4 TMCC IDs
            for i in range(4):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL4))
        elif self._mode == 3:
            # Switch mode, latched, 4 TMCC IDs
            for i in range(4):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL5))
        return cmds


class Bpc2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Bpc2Req) -> None:
        super().__init__(PdiDevice.BPC2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # TR mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Bpc2Req(self.tmcc_id + i, action=Bpc2Action.CONTROL1))
        elif self._mode == 1:
            # TR mode, 1 TMCC ID
            cmds.append(Bpc2Req(self.tmcc_id, action=Bpc2Action.CONTROL2))
        elif self._mode == 2:
            # Acc mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Bpc2Req(self.tmcc_id + i, action=Bpc2Action.CONTROL3))
        elif self._mode == 3:
            # Acc mode, 1 TMCC ID
            cmds.append(Bpc2Req(self.tmcc_id, action=Bpc2Action.CONTROL4))
        return cmds


class Stm2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Stm2Req) -> None:
        super().__init__(PdiDevice.STM2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # 16 inputs, 16 TMCC IDs
            for i in range(16):
                cmds.append(Stm2Req(self.tmcc_id + i, action=Stm2Action.CONTROL1))
        elif self._mode == 2:
            # 8 input pairs, 8 TMCC IDs
            for i in range(8):
                cmds.append(Stm2Req(self.tmcc_id + i, action=Stm2Action.CONTROL1))
        return cmds


class IrdaDeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: IrdaReq) -> None:
        super().__init__(PdiDevice.IRDA, cmd)

    @property
    def state_requests(self) -> List[T]:
        cmds = [
            IrdaReq(self.tmcc_id, action=IrdaAction.INFO, scope=CommandScope.ACC),
            IrdaReq(self.tmcc_id, action=IrdaAction.CONFIG),
        ]
        return cmds


A = TypeVar("A", bound=PdiAction)


class DeviceWrapper:
    C = TypeVar("C", bound=PdiReq.__class__)
    E = TypeVar("E", bound=Enum)
    T = TypeVar("T", bound=PdiReq)
    DC = TypeVar("DC", bound=PdiDeviceConfig.__class__)

    def __init__(
        self,
        req_class: C,
        *commands: PdiCommand,
        enums: E = None,
        dev_class: DC = None,
        common_actions: bool = True,
    ) -> None:
        self.req_class = req_class
        self.enums = enums
        self.dev_class = dev_class
        self.commands = commands
        self.common_actions = common_actions
        self.get: PdiCommand = self._harvest_command("GET")
        self.set: PdiCommand = self._harvest_command("SET")
        self.rx: PdiCommand = self._harvest_command("RX")

    def __repr__(self) -> str:
        return f"{self.req_class.__class__.__name__}"

    def build(self, data: bytes | int, action: A = None) -> T:
        if isinstance(data, bytes):
            if self.enums is not None:
                action_byte = data[3] if len(data) > 3 else None
                if action_byte & 0x80 == 0x80:
                    error = True
                    action_byte = 0x7FF & action_byte
                else:
                    error = False
                action = self.enums.by_value(action_byte)
                return self.req_class(data, action=action, error=error)
            else:
                return self.req_class(data)
        else:
            return self.req_class(data, action=action)

    def firmware(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("FIRMWARE")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def status(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("STATUS")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def info(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("INFO")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def config(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("CONFIG")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def clear_errors(self, tmcc_id: int) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("CLEAR_ERRORS")
            return self.req_class(tmcc_id, self.set, enum)
        return None

    def reset(self, tmcc_id: int) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("RESET")
            return self.req_class(tmcc_id, self.set, enum)
        return None

    def identify(self, tmcc_id: int, ident: int = 1) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("IDENTIFY")
            return self.req_class(tmcc_id, self.set, enum, ident=ident)
        return None

    @property
    def is_common_actions(self) -> bool:
        return self.common_actions

    def _harvest_command(self, suffix: str) -> PdiCommand | None:
        suffix = suffix.strip().upper()
        for e in self.commands:
            if e.name.endswith(suffix):
                return e
        return None


D = TypeVar("D", bound=PdiDeviceConfig)


@unique
class PdiDevice(Mixins, FriendlyMixins):
    """
    All supported LCS/PDI devices should be listed here, along with their
    constructor class wrapped in a DeviceWrapper
    """

    from .asc2_req import Asc2Req
    from .bpc2_req import Bpc2Req
    from .wifi_req import WiFiReq
    from .pdi_req import AllReq, PingReq, TmccReq
    from .base_req import BaseReq
    from .d4_req import D4Req
    from .stm2_req import Stm2Req
    from .irda_req import IrdaReq
    from .lcs_req import Ser2Req
    from .block_req import BlockReq

    BASE = DeviceWrapper(BaseReq)
    D4 = DeviceWrapper(D4Req, enums=D4Action, common_actions=False)
    PING = DeviceWrapper(PingReq)
    ALL = DeviceWrapper(AllReq, PdiCommand.ALL_GET, PdiCommand.ALL_SET)
    TMCC = DeviceWrapper(TmccReq, PdiCommand.TMCC_TX, PdiCommand.TMCC_RX)
    TMCC4 = DeviceWrapper(TmccReq, PdiCommand.TMCC4_TX, PdiCommand.TMCC4_RX)
    WIFI = DeviceWrapper(WiFiReq, PdiCommand.WIFI_GET, PdiCommand.WIFI_SET, PdiCommand.WIFI_RX, enums=WiFiAction)
    SER2 = DeviceWrapper(Ser2Req, PdiCommand.SER2_GET, PdiCommand.SER2_SET, PdiCommand.SER2_RX, enums=Ser2Action)
    BLOCK = DeviceWrapper(BlockReq, PdiCommand.BLOCK_GET, PdiCommand.BLOCK_SET, PdiCommand.BLOCK_RX)
    IRDA = DeviceWrapper(
        IrdaReq,
        PdiCommand.IRDA_GET,
        PdiCommand.IRDA_SET,
        PdiCommand.IRDA_RX,
        enums=IrdaAction,
        dev_class=IrdaDeviceConfig,
    )
    ASC2 = DeviceWrapper(
        Asc2Req,
        PdiCommand.ASC2_GET,
        PdiCommand.ASC2_SET,
        PdiCommand.ASC2_RX,
        enums=Asc2Action,
        dev_class=Acs2DeviceConfig,
    )
    BPC2 = DeviceWrapper(
        Bpc2Req,
        PdiCommand.BPC2_GET,
        PdiCommand.BPC2_SET,
        PdiCommand.BPC2_RX,
        enums=Bpc2Action,
        dev_class=Bpc2DeviceConfig,
    )
    STM2 = DeviceWrapper(
        Stm2Req,
        PdiCommand.STM2_GET,
        PdiCommand.STM2_SET,
        PdiCommand.STM2_RX,
        enums=Stm2Action,
        dev_class=Stm2DeviceConfig,
    )
    UPDATE = DeviceWrapper(BaseReq)

    @classmethod
    def from_pdi_command(cls, cmd: PdiCommand) -> PdiDevice:
        return cls(cmd.name.split("_")[0].upper())

    @classmethod
    def from_data(cls, data: bytes) -> PdiDevice:
        return cls(PdiCommand(data[1]).name.split("_")[0].upper())

    @property
    def can_build_device(self) -> bool:
        return self.value.dev_class is not None

    def build_req(self, data: bytes | int, action: A = None) -> T:
        return self.value.build(data, action)

    def build_device(self, cmd: T) -> D:
        return self.value.dev_class(cmd)

    def firmware(self, tmcc_id: int) -> T:
        return self.value.firmware(tmcc_id)

    def config(self, tmcc_id: int) -> T:
        return self.value.config(tmcc_id)

    def status(self, tmcc_id: int) -> T:
        return self.value.status(tmcc_id)

    def info(self, tmcc_id: int) -> T:
        return self.value.info(tmcc_id)

    def clear_errors(self, tmcc_id: int) -> T:
        """
        Build a Clear Errors request
        """
        return self.value.clear_errors(tmcc_id)

    def reset(self, tmcc_id: int) -> T:
        """
        Build a Reset request
        """
        return self.value.identify(tmcc_id)

    def identify(self, tmcc_id: int, ident: int = 1) -> T:
        """
        Build an Identify request
        """
        return self.value.identify(tmcc_id, ident)


class SystemDeviceDict(defaultdict):
    """
    Maintains a dictionary of CommandScope to ComponentStateDict
    """

    def __missing__(self, key: PdiDevice) -> PdiDeviceDict:
        """
        generate a ComponentState object for the dictionary, based on the key
        """
        if isinstance(key, PdiDevice):
            device = key
        else:
            raise KeyError(f"Invalid device: {key}")
        # create the component state dict for this key
        self[key] = PdiDeviceDict(device)
        return self[key]

    def register_pdi_device(self, cmd: PdiReq) -> List[T] | None:
        if cmd.action.bits == CommonAction.CONFIG.bits:
            tmcc_id = cmd.tmcc_id
            device = cmd.pdi_device
            if device.can_build_device is True:
                dev_config = device.build_device(cmd)
                self[device][tmcc_id] = dev_config
                return dev_config.state_requests
        return None


class PdiDeviceDict(dict):
    def __init__(self, device: PdiDevice):
        super().__init__()  # base class doesn't get a factory
        self._device = device

    @property
    def device(self) -> PdiDevice:
        return self._device



from __future__ import annotations

import argparse
import sys
from collections.abc import Sequence
from time import sleep
from typing import List, Callable, TypeVar, Tuple

from ..multibyte.multibyte_constants import TMCC2RailSoundsDialogControl

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from ..command_def import CommandDefEnum
from ..command_req import CommandReq
from ..constants import (
    DEFAULT_ADDRESS,
    DEFAULT_BAUDRATE,
    DEFAULT_PORT,
    OfficialRRSpeeds,
)
from ..constants import CommandScope
from ..tmcc1.tmcc1_constants import TMCC1EngineCommandEnum, TMCC1RRSpeedsEnum
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum, TMCC2RRSpeedsEnum

from ...comm.comm_buffer import CommBuffer
from ...utils.argument_parser import ArgumentParser

T = TypeVar("T", TMCC1RRSpeedsEnum, TMCC2RRSpeedsEnum)


class SequenceReq(CommandReq, Sequence):
    from .sequence_constants import SequenceCommandEnum

    @classmethod
    def build(
        cls,
        command: SequenceCommandEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
    ) -> Self:
        if command.value.cmd_class is None:
            cls._register_command_class(command)
        cmd_class = command.value.cmd_class
        return cmd_class(address, data, scope)

    @classmethod
    def _register_command_class(cls, command):
        """
        We need this function to avoid Python circular dependencies
        """
        from .sequence_constants import SequenceCommandEnum
        from .speed_req import SpeedReq
        from .abs_speed_rpm import AbsoluteSpeedRpm
        from .grade_crossing_req import GradeCrossingReq
        from .ramped_speed_req import RampedSpeedReq
        from .ramped_speed_req import RampedSpeedDialogReq
        from .labor_effect import LaborEffectUpReq
        from .labor_effect import LaborEffectDownReq

        if command == SequenceCommandEnum.ABSOLUTE_SPEED_SEQ:
            SequenceCommandEnum.ABSOLUTE_SPEED_SEQ.value.register_cmd_class(SpeedReq)
        elif command == SequenceCommandEnum.ABSOLUTE_SPEED_RPM:
            SequenceCommandEnum.ABSOLUTE_SPEED_RPM.value.register_cmd_class(AbsoluteSpeedRpm)
        elif command == SequenceCommandEnum.GRADE_CROSSING_SEQ:
            SequenceCommandEnum.GRADE_CROSSING_SEQ.value.register_cmd_class(GradeCrossingReq)
        elif command == SequenceCommandEnum.RAMPED_SPEED_SEQ:
            SequenceCommandEnum.RAMPED_SPEED_SEQ.value.register_cmd_class(RampedSpeedReq)
        elif command == SequenceCommandEnum.RAMPED_SPEED_DIALOG_SEQ:
            SequenceCommandEnum.RAMPED_SPEED_DIALOG_SEQ.value.register_cmd_class(RampedSpeedDialogReq)
        elif command == SequenceCommandEnum.LABOR_EFFECT_UP_SEQ:
            SequenceCommandEnum.LABOR_EFFECT_UP_SEQ.value.register_cmd_class(LaborEffectUpReq)
        elif command == SequenceCommandEnum.LABOR_EFFECT_DOWN_SEQ:
            SequenceCommandEnum.LABOR_EFFECT_DOWN_SEQ.value.register_cmd_class(LaborEffectDownReq)

    def __init__(
        self,
        command: SequenceCommandEnum,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = None,
        repeat: int = 1,
        delay: float = 0,
        requests: List[CommandReq] = None,
    ) -> None:
        self._requests: List[SequencedReq] = list()  # need to define prior to calling super()
        if requests:
            for request in requests:
                self._requests.append(SequencedReq(request, repeat, delay))
        super().__init__(command, address, 0, scope)
        self._repeat = repeat
        self._delay = delay

    def __getitem__(self, index) -> SequencedReq:
        return self._requests[index]

    @property
    def as_bytes(self) -> bytes:
        self._recalculate()
        cmd_bytes = bytes()
        for req_wrapper in self._requests:
            req = req_wrapper.request
            cmd_bytes += req.as_bytes
        return cmd_bytes

    @property
    def requests(self) -> List[SequencedReq]:
        return self._requests.copy()

    def __len__(self) -> int:
        return len(self._requests)

    def _apply_address(self, new_address: int = None) -> int:
        for req_wrapper in self._requests:
            req = req_wrapper.request
            req.address = self.address
        return 0

    def _apply_data(self, new_data: int = None) -> int:
        for req_wrapper in self._requests:
            req = req_wrapper.request
            if req.is_data:
                req.data = self.data
        return 0

    def add(
        self,
        request: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        repeat: int = 1,
        delay: float = 0,
    ) -> None:
        if isinstance(request, CommandDefEnum):
            request = CommandReq.build(request, address=address, data=data, scope=scope)
        self._requests.append(SequencedReq(request, repeat=repeat, delay=delay))

    def send(
        self,
        repeat: int = None,
        delay: float = None,
        duration: float = None,
        interval: int = None,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        for sqr in self._requests:
            request = sqr.request
            req_repeat = sqr.repeat if sqr.repeat is not None else repeat
            req_delay = sqr.delay if sqr.delay is not None else delay
            req_duration = sqr.duration if sqr.duration is not None else duration
            req_interval = sqr.interval if sqr.interval is not None else interval
            request.send(
                repeat=req_repeat,
                delay=req_delay,
                duration=req_duration,
                interval=req_interval,
                baudrate=baudrate,
                port=port,
                server=server,
            )
            sleep(0.001)

    def fire(
        self,
        repeat: int = None,
        delay: float = None,
        duration: float = None,
        interval: int = None,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        self.send(
            repeat=repeat,
            delay=delay,
            duration=duration,
            interval=interval,
            baudrate=baudrate,
            port=port,
            server=server,
        )
        CommBuffer.build().shutdown()

    def as_action(
        self,
        repeat: int = 1,
        delay: float = 0,
        duration: float = 0,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
        address: int = None,
        data: int = None,
    ) -> Callable:
        buffer = CommBuffer.build(baudrate=baudrate, port=port, server=server)

        def send_func(new_address: int = None, new_data: int = None) -> None:
            if new_address and new_address != self.address:
                self.address = new_address
            if new_data != self.data:
                self.data = new_data
            elif address and address != self.address:
                self.address = address
            if self.num_data_bits:
                if new_data is not None and new_data != self.data:
                    self.data = new_data
                elif data is not None and data != self.data:
                    self.data = data
            for sq_request in self._requests:
                request = sq_request.request
                request._enqueue_command(
                    request.as_bytes,
                    repeat=sq_request.repeat,
                    delay=sq_request.delay,
                    duration=sq_request.duration,
                    interval=sq_request.interval,
                    baudrate=baudrate,
                    port=port,
                    server=server,
                    buffer=buffer,
                )

        return send_func

    @staticmethod
    def speed_parser(is_tmcc: bool = False) -> ArgumentParser:
        """
        Parse the first token of the user's input
        """
        cde = TMCC1EngineCommandEnum if is_tmcc else TMCC2EngineCommandEnum
        command_parser = ArgumentParser(exit_on_error=False)
        group = command_parser.add_mutually_exclusive_group()
        group.add_argument("-stop", action="store_const", const=cde.SPEED_STOP_HOLD, dest="command")
        group.add_argument("-roll", action="store_const", const=cde.SPEED_ROLL, dest="command")
        group.add_argument("-restricted", action="store_const", const=cde.SPEED_RESTRICTED, dest="command")
        group.add_argument("-slow", action="store_const", const=cde.SPEED_SLOW, dest="command")
        group.add_argument("-medium", action="store_const", const=cde.SPEED_MEDIUM, dest="command")
        group.add_argument("-limited", action="store_const", const=cde.SPEED_LIMITED, dest="command")
        group.add_argument("-normal", action="store_const", const=cde.SPEED_NORMAL, dest="command")
        group.add_argument("-hi", "-high", "-highball", action="store_const", const=cde.SPEED_HIGHBALL, dest="command")
        return command_parser

    def decode_rr_speed(
        self, speed: OfficialRRSpeeds | int | str, is_tmcc: bool
    ) -> Tuple[CommandDefEnum, CommandDefEnum, int, CommandDefEnum]:
        base = None
        speed_enum = None
        speed_int = None
        tower = engr = None
        if isinstance(speed, OfficialRRSpeeds):
            base = f"SPEED_{speed.name}"
            if isinstance(speed, TMCC1RRSpeedsEnum):
                speed_enum = TMCC1EngineCommandEnum.by_name(base)
            else:
                speed_enum = TMCC2EngineCommandEnum.by_name(base)
            if speed_enum is None:
                raise ValueError(f"Unknown speed type: {speed}")
        elif isinstance(speed, int):
            if is_tmcc:
                for rr_speed in TMCC1RRSpeedsEnum:
                    if speed in rr_speed.value:
                        speed_int = rr_speed.value[0]
                        base = f"SPEED_{rr_speed.name}"
                        speed_enum = TMCC1EngineCommandEnum.by_name(base)
                        break
            else:
                for rr_speed in TMCC2RRSpeedsEnum:
                    if speed in rr_speed.value:
                        speed_int = rr_speed.value[0]
                        base = f"SPEED_{rr_speed.name}"
                        speed_enum = TMCC2EngineCommandEnum.by_name(base)
                        break
        elif isinstance(speed, str):
            try:
                args = self.speed_parser().parse_args(["-" + speed.strip()])
                speed_enum = args.command
                base = speed_enum.name
                _, speed_int = speed_enum.value.alias
            except argparse.ArgumentError:
                pass

        if base is not None:
            tower = TMCC2RailSoundsDialogControl.by_name(f"TOWER_{base}")
            engr = TMCC2RailSoundsDialogControl.by_name(f"ENGINEER_{base}")
        return tower, speed_enum, speed_int, engr

    def _recalculate(self) -> None:
        """
        Recalculate command state, prior to sending bytes
        """
        pass


class SequencedReq:
    def __init__(
        self,
        request: CommandReq,
        repeat: int = None,
        delay: float = None,
        duration: float = None,
        interval: int = None,
    ) -> None:
        self.request: CommandReq = request
        self.repeat: int = repeat
        self.delay: float = delay
        self.duration: float = duration
        self.interval: int = interval

    def __repr__(self) -> str:
        rtn = f"<{self.request} repeat: {self.repeat} delay: {self.delay} "
        rtn += f"duration: {self.duration} interval: {self.interval}>"
        return rtn

    @property
    def as_bytes(self) -> bytes:
        return self.request.as_bytes
